#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

import sys

from functools import reduce
import warnings
import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf import gto
from pyscf import scf
from pyscf import ao2mo
from pyscf import fci
from pyscf.mcscf import addons
from pyscf import __config__

WITH_META_LOWDIN = getattr(__config__, 'mcscf_analyze_with_meta_lowdin', True)
LARGE_CI_TOL = getattr(__config__, 'mcscf_analyze_large_ci_tol', 0.1)
PENALTY = getattr(__config__, 'mcscf_casci_CASCI_fix_spin_shift', 0.2)
FRAC_OCC_THRESHOLD = 1e-6

if sys.version_info < (3,):
    RANGE_TYPE = list
else:
    RANGE_TYPE = range


def h1e_for_cas(casci, mo_coeff=None, ncas=None, ncore=None):
    '''CAS space one-electron hamiltonian

    Args:
        casci : a CASSCF/CASCI object or RHF object

    Returns:
        A tuple, the first is the effective one-electron hamiltonian defined in CAS space,
        the second is the electronic energy from core.
    '''
    if mo_coeff is None: mo_coeff = casci.mo_coeff
    if ncas is None: ncas = casci.ncas
    if ncore is None: ncore = casci.ncore
    mo_core = mo_coeff[:,:ncore]
    mo_cas = mo_coeff[:,ncore:ncore+ncas]

    hcore = casci.get_hcore()
    energy_core = casci.energy_nuc()
    if mo_core.size == 0:
        corevhf = 0
    else:
        core_dm = numpy.dot(mo_core, mo_core.conj().T) * 2
        corevhf = casci.get_veff(casci.mol, core_dm)
        energy_core += numpy.einsum('ij,ji', core_dm, hcore).real
        energy_core += numpy.einsum('ij,ji', core_dm, corevhf).real * .5
    h1eff = reduce(numpy.dot, (mo_cas.conj().T, hcore+corevhf, mo_cas))
    return h1eff, energy_core

def analyze(casscf, mo_coeff=None, ci=None, verbose=None,
            large_ci_tol=LARGE_CI_TOL, with_meta_lowdin=WITH_META_LOWDIN,
            **kwargs):
    from pyscf.lo import orth
    from pyscf.tools import dump_mat
    from pyscf.mcscf import addons
    log = logger.new_logger(casscf, verbose)

    if mo_coeff is None: mo_coeff = casscf.mo_coeff
    if ci is None: ci = casscf.ci
    nelecas = casscf.nelecas
    ncas = casscf.ncas
    ncore = casscf.ncore
    nocc = ncore + ncas
    mocore = mo_coeff[:,:ncore]
    mocas = mo_coeff[:,ncore:nocc]

    label = casscf.mol.ao_labels()
    if (isinstance(ci, (list, tuple, RANGE_TYPE)) and
        not isinstance(casscf.fcisolver, addons.StateAverageFCISolver)):
        log.warn('Mulitple states found in CASCI/CASSCF solver. Density '
                 'matrix of the first state is generated in .analyze() function.')
        civec = ci[0]
    else:
        civec = ci
    if getattr(casscf.fcisolver, 'make_rdm1s', None):
        casdm1a, casdm1b = casscf.fcisolver.make_rdm1s(civec, ncas, nelecas)
        casdm1 = casdm1a + casdm1b
        dm1b = numpy.dot(mocore, mocore.conj().T)
        dm1a = dm1b + reduce(numpy.dot, (mocas, casdm1a, mocas.conj().T))
        dm1b += reduce(numpy.dot, (mocas, casdm1b, mocas.conj().T))
        dm1 = dm1a + dm1b
        spin_dm1 = dm1a - dm1b
        if log.verbose >= logger.DEBUG2:
            log.info('alpha density matrix (on AO)')
            dump_mat.dump_tri(log.stdout, dm1a, label, **kwargs)
            log.info('beta density matrix (on AO)')
            dump_mat.dump_tri(log.stdout, dm1b, label, **kwargs)
    else:
        casdm1 = casscf.fcisolver.make_rdm1(civec, ncas, nelecas)
        dm1a = (numpy.dot(mocore, mocore.conj().T) * 2 +
                reduce(numpy.dot, (mocas, casdm1, mocas.conj().T)))
        dm1b = None
        dm1 = dm1a
        spin_dm1 = None

    if log.verbose >= logger.INFO:
        ovlp_ao = casscf._scf.get_ovlp()
        # note the last two args of ._eig for mc1step_symm
        occ, ucas = casscf._eig(-casdm1, ncore, nocc)
        log.info('Natural occ %s', str(-occ))
        mocas = numpy.dot(mocas, ucas)
        if with_meta_lowdin:
            log.info('Natural orbital (expansion on meta-Lowdin AOs) in CAS space')
            orth_coeff = orth.orth_ao(casscf.mol, 'meta_lowdin', s=ovlp_ao)
            mocas = reduce(numpy.dot, (orth_coeff.conj().T, ovlp_ao, mocas))
        else:
            log.info('Natural orbital (expansion on AOs) in CAS space')
        dump_mat.dump_rec(log.stdout, mocas, label, start=1, **kwargs)
        if log.verbose >= logger.DEBUG2:
            if not casscf.natorb:
                log.debug2('NOTE: mc.mo_coeff in active space is different to '
                           'the natural orbital coefficients printed in above.')
            if with_meta_lowdin:
                c = reduce(numpy.dot, (orth_coeff.conj().T, ovlp_ao, mo_coeff))
                log.debug2('MCSCF orbital (expansion on meta-Lowdin AOs)')
            else:
                c = mo_coeff
                log.debug2('MCSCF orbital (expansion on AOs)')
            dump_mat.dump_rec(log.stdout, c, label, start=1, **kwargs)

        if casscf._scf.mo_coeff is not None:
            addons.map2hf(casscf, casscf._scf.mo_coeff)

        if (ci is not None and
            (getattr(casscf.fcisolver, 'large_ci', None) or
             getattr(casscf.fcisolver, 'states_large_ci', None))):
            log.info('** Largest CI components **')
            if isinstance(ci, (list, tuple, RANGE_TYPE)):
                if hasattr(casscf.fcisolver, 'states_large_ci'):
                    # defined in state_average_mix_ mcscf object
                    res = casscf.fcisolver.states_large_ci(ci, casscf.ncas, casscf.nelecas,
                                                           large_ci_tol, return_strs=False)
                else:
                    res = [casscf.fcisolver.large_ci(civec, casscf.ncas, casscf.nelecas,
                                                     large_ci_tol, return_strs=False)
                           for civec in ci]
                for i, civec in enumerate(ci):
                    log.info('  [alpha occ-orbitals] [beta occ-orbitals]  state %-3d CI coefficient', i)
                    for c,ia,ib in res[i]:
                        log.info('  %-20s %-30s % .12f', ia, ib, c)
            else:
                log.info('  [alpha occ-orbitals] [beta occ-orbitals]            CI coefficient')
                res = casscf.fcisolver.large_ci(ci, casscf.ncas, casscf.nelecas,
                                                large_ci_tol, return_strs=False)
                for c,ia,ib in res:
                    log.info('  %-20s %-30s % .12f', ia, ib, c)

        if with_meta_lowdin:
            casscf._scf.mulliken_meta(casscf.mol, dm1, s=ovlp_ao, verbose=log)
        else:
            casscf._scf.mulliken_pop(casscf.mol, dm1, s=ovlp_ao, verbose=log)
        if spin_dm1 is not None:
            if with_meta_lowdin:
                log.info('Mulliken spin population analysis on meta-Lowdin AOs:')
                spin_pop, spin_chg = casscf._scf.mulliken_meta(casscf.mol, spin_dm1, s=ovlp_ao, verbose=log)
            else:
                log.info('Mulliken spin population analysis on AOs:')
                spin_pop, spin_chg = casscf._scf.mulliken_pop(casscf.mol, spin_dm1, s=ovlp_ao, verbose=log)
            for i, s in enumerate(label):
                log.info('spop of %-12s %10.5f', s, spin_pop[i])
            spin_chg = casscf.mol.atom_charges() - spin_chg
            log.note('Mulliken atomic spins:')
            for ia in range(casscf.mol.natm):
                symb = casscf.mol.atom_symbol(ia)
                log.note('spin of  %d%s =   %10.5f', ia, symb, spin_chg[ia])
    return dm1a, dm1b

def get_fock(mc, mo_coeff=None, ci=None, eris=None, casdm1=None, verbose=None):
    r'''
    Effective one-electron Fock matrix in AO representation
    f = \sum_{pq} E_{pq} F_{pq}
    F_{pq} = h_{pq} + \sum_{rs} [(pq|rs)-(ps|rq)] DM_{sr}

    Ref.
    Theor. Chim. Acta., 91, 31
    Chem. Phys. 48, 157

    For state-average CASCI/CASSCF object, the effective fock matrix is based
    on the state-average density matrix.  To obtain Fock matrix of a specific
    state in the state-average calculations, you can pass "casdm1" of the
    specific state to this function.

    Args:
        mc: a CASSCF/CASCI object or RHF object

    Kwargs:
        mo_coeff (ndarray): orbitals that span the core, active and external
            space.
        ci (ndarray): CI coefficients (or objects to represent the CI
            wavefunctions in DMRG/QMC-MCSCF calculations).
        eris: Integrals for the MCSCF object. Input this object to reduce the
            overhead of computing integrals. It can be generated by
            :func:`mc.ao2mo` method.
        casdm1 (ndarray): 1-particle density matrix in active space. Without
            input casdm1, the density matrix is computed with the input ci
            coefficients/object. If neither ci nor casdm1 were given, density
            matrix is computed by :func:`mc.fcisolver.make_rdm1` method. For
            state-average CASCI/CASCF calculation, this results in the
            effective Fock matrix based on the state-average density matrix.
            To obtain the effective Fock matrix for one particular state, you
            can assign the density matrix of that state to the kwarg casdm1.

    Returns:
        Fock matrix
    '''

    if ci is None: ci = mc.ci
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    nmo = mo_coeff.shape[1]
    ncore = mc.ncore
    ncas = mc.ncas
    nocc = ncore + ncas
    nelecas = mc.nelecas

    if casdm1 is None:
        casdm1 = mc.fcisolver.make_rdm1(ci, ncas, nelecas)
    if getattr(eris, 'ppaa', None) is not None:
        vj = numpy.empty((nmo,nmo))
        vk = numpy.empty((nmo,nmo))
        for i in range(nmo):
            vj[i] = numpy.einsum('ij,qij->q', casdm1, eris.ppaa[i])
            vk[i] = numpy.einsum('ij,iqj->q', casdm1, eris.papa[i])
        mo_inv = numpy.dot(mo_coeff.conj().T, mc._scf.get_ovlp())
        fock = (mc.get_hcore() +
                reduce(numpy.dot, (mo_inv.conj().T, eris.vhf_c+vj-vk*.5, mo_inv)))
    else:
        dm_core = numpy.dot(mo_coeff[:,:ncore]*2, mo_coeff[:,:ncore].conj().T)
        mocas = mo_coeff[:,ncore:nocc]
        dm = dm_core + reduce(numpy.dot, (mocas, casdm1, mocas.conj().T))
        vj, vk = mc._scf.get_jk(mc.mol, dm)
        fock = mc.get_hcore() + vj-vk*.5
    return fock

def cas_natorb(mc, mo_coeff=None, ci=None, eris=None, sort=False,
               casdm1=None, verbose=None, with_meta_lowdin=WITH_META_LOWDIN):
    '''Transform active orbitals to natural orbitals, and update the CI wfn
    accordingly

    Args:
        mc : a CASSCF/CASCI object or RHF object

    Kwargs:
        sort : bool
            Sort natural orbitals wrt the occupancy.

    Returns:
        A tuple, the first item is natural orbitals, the second is updated CI
        coefficients, the third is the natural occupancy associated to the
        natural orbitals.
    '''
    from pyscf.lo import orth
    from pyscf.tools import dump_mat
    from pyscf.tools.mo_mapping import mo_1to1map
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    log = logger.new_logger(mc, verbose)
    ncore = mc.ncore
    ncas = mc.ncas
    nocc = ncore + ncas
    nelecas = mc.nelecas
    nmo = mo_coeff.shape[1]
    if casdm1 is None:
        casdm1 = mc.fcisolver.make_rdm1(ci, ncas, nelecas)
    if getattr(mo_coeff, 'orbsym', None) is not None:
        orbsym = numpy.copy(mo_coeff.orbsym)
    else:
        orbsym = numpy.zeros(mo_coeff.shape[1], dtype=int)
    if getattr(mc, 'extrasym', None) is not None:
        orbsym_extra = numpy.asarray([str(i1) + str(i2)
                                      for i1, i2 in zip(orbsym, mc.extrasym)])
    else:
        orbsym_extra = orbsym
    # orbital symmetry is reserved in this _eig call
    cas_occ, ucas = mc._eig(-casdm1, ncore, nocc, orbsym_extra[ncore:nocc])
    if sort:
        casorb_idx = numpy.argsort(cas_occ.round(9), kind='mergesort')
        cas_occ = cas_occ[casorb_idx]
        ucas = ucas[:,casorb_idx]

    cas_occ = -cas_occ
    mo_occ = numpy.zeros(mo_coeff.shape[1])
    mo_occ[:ncore] = 2
    mo_occ[ncore:nocc] = cas_occ

    mo_coeff1 = mo_coeff.copy()
    mo_coeff1[:,ncore:nocc] = numpy.dot(mo_coeff[:,ncore:nocc], ucas)
    if getattr(mo_coeff, 'orbsym', None) is not None:
        if sort:
            orbsym[ncore:nocc] = orbsym[ncore:nocc][casorb_idx]
        mo_coeff1 = lib.tag_array(mo_coeff1, orbsym=orbsym)
    else:
        orbsym = numpy.zeros(nmo, dtype=int)

    # When occupancies of active orbitals equal to 2 or 0, these orbitals
    # need to be canonicalized along with inactive(core or virtual) orbitals
    # using general Fock matrix. Because they are strongly coupled with
    # inactive orbitals, the 0th order Hamiltonian of MRPT methods can be
    # strongly affected. Numerical uncertainty may be found in the perturbed
    # correlation energy.
    occ2_idx = numpy.where(2 - cas_occ < FRAC_OCC_THRESHOLD)[0]
    occ0_idx = numpy.where(cas_occ < FRAC_OCC_THRESHOLD)[0]
    if occ2_idx.size > 0 or occ0_idx.size > 0:
        fock_ao = mc.get_fock(mo_coeff, ci, eris, casdm1, verbose)

        def _diag_subfock_(idx):
            c = mo_coeff1[:,idx]
            fock = reduce(numpy.dot, (c.conj().T, fock_ao, c))
            w, c = mc._eig(fock, None, None, orbsym[idx])
            mo_coeff1[:,idx] = mo_coeff1[:,idx].dot(c)

        if occ2_idx.size > 0:
            log.warn('Active orbitals %s (occs = %s) are canonicalized with core orbitals',
                     occ2_idx, cas_occ[occ2_idx])
            full_occ2_idx = numpy.append(numpy.arange(ncore), ncore + occ2_idx)
            _diag_subfock_(full_occ2_idx)
        if occ0_idx.size > 0:
            log.warn('Active orbitals %s (occs = %s) are canonicalized with external orbitals',
                     occ0_idx, cas_occ[occ0_idx])
            full_occ0_idx = numpy.append(ncore + occ0_idx, numpy.arange(nocc, nmo))
            _diag_subfock_(full_occ0_idx)

    # Rotate CI according to the unitary coefficients ucas if applicable
    fcivec = None
    if getattr(mc.fcisolver, 'transform_ci_for_orbital_rotation', None):
        if isinstance(ci, (fci.FCIvector, fci.SCIvector, numpy.ndarray)):
            fcivec = mc.fcisolver.transform_ci_for_orbital_rotation(ci, ncas, nelecas, ucas)
        elif (isinstance(ci, (list, tuple)) and
              all(isinstance(x[0], (fci.FCIvector, fci.SCIvector, numpy.ndarray)) for x in ci)):
            fcivec = [mc.fcisolver.transform_ci_for_orbital_rotation(x, ncas, nelecas, ucas)
                      for x in ci]
    elif getattr(mc.fcisolver, 'states_transform_ci_for_orbital_rotation', None):
        fcivec = mc.fcisolver.states_transform_ci_for_orbital_rotation(ci, ncas, nelecas, ucas)

    # Rerun fcisolver to get wavefunction if it cannot be transformed from
    # existed one.
    if fcivec is None:
        log.info('FCI vector not available, call CASCI to update wavefunction')
        mocas = mo_coeff1[:,ncore:nocc]
        hcore = mc.get_hcore()
        dm_core = numpy.dot(mo_coeff1[:,:ncore]*2, mo_coeff1[:,:ncore].conj().T)
        ecore = mc.energy_nuc()
        ecore+= numpy.einsum('ij,ji', hcore, dm_core)
        h1eff = reduce(numpy.dot, (mocas.conj().T, hcore, mocas))
        if getattr(eris, 'ppaa', None) is not None:
            ecore += eris.vhf_c[:ncore,:ncore].trace()
            h1eff += reduce(numpy.dot, (ucas.conj().T, eris.vhf_c[ncore:nocc,ncore:nocc], ucas))
            aaaa = ao2mo.restore(4, eris.ppaa[ncore:nocc,ncore:nocc,:,:], ncas)
            aaaa = ao2mo.incore.full(aaaa, ucas)
        else:
            if getattr(mc, 'with_df', None):
                aaaa = mc.with_df.ao2mo(mocas)
            else:
                aaaa = ao2mo.kernel(mc.mol, mocas)
            corevhf = mc.get_veff(mc.mol, dm_core)
            ecore += numpy.einsum('ij,ji', dm_core, corevhf) * .5
            h1eff += reduce(numpy.dot, (mocas.conj().T, corevhf, mocas))


        # See label_symmetry_ function in casci_symm.py which initialize the
        # orbital symmetry information in fcisolver.  This orbital symmetry
        # labels should be reordered to match the sorted active space orbitals.
        if sort and getattr(mo_coeff1, 'orbsym', None) is not None:
            mc.fcisolver.orbsym = mo_coeff1.orbsym[ncore:nocc]

        max_memory = max(400, mc.max_memory-lib.current_memory()[0])
        e, fcivec = mc.fcisolver.kernel(h1eff, aaaa, ncas, nelecas, ecore=ecore,
                                        max_memory=max_memory, verbose=log)
        log.debug('In Natural orbital, CASCI energy = %s', e)

    if log.verbose >= logger.INFO:
        ovlp_ao = mc._scf.get_ovlp()
        # where_natorb gives the new locations of the natural orbitals
        where_natorb = mo_1to1map(ucas)
        log.debug('where_natorb %s', str(where_natorb))
        log.info('Natural occ %s', str(cas_occ))
        if with_meta_lowdin:
            log.info('Natural orbital (expansion on meta-Lowdin AOs) in CAS space')
            label = mc.mol.ao_labels()
            orth_coeff = orth.orth_ao(mc.mol, 'meta_lowdin', s=ovlp_ao)
            mo_cas = reduce(numpy.dot, (orth_coeff.conj().T, ovlp_ao, mo_coeff1[:,ncore:nocc]))
        else:
            log.info('Natural orbital (expansion on AOs) in CAS space')
            label = mc.mol.ao_labels()
            mo_cas = mo_coeff1[:,ncore:nocc]
        dump_mat.dump_rec(log.stdout, mo_cas, label, start=1)

        if mc._scf.mo_coeff is not None:
            s = reduce(numpy.dot, (mo_coeff1[:,ncore:nocc].conj().T,
                                   mc._scf.get_ovlp(), mc._scf.mo_coeff))
            idx = numpy.argwhere(abs(s)>.4)
            for i,j in idx:
                log.info('<CAS-nat-orb|mo-hf>  %-5d  %-5d  % 12.8f',
                         ncore+i+1, j+1, s[i,j])
    return mo_coeff1, fcivec, mo_occ

def canonicalize(mc, mo_coeff=None, ci=None, eris=None, sort=False,
                 cas_natorb=False, casdm1=None, verbose=logger.NOTE,
                 with_meta_lowdin=WITH_META_LOWDIN, stav_dm1=False):
    '''Canonicalized CASCI/CASSCF orbitals of effective Fock matrix and
    update CI coefficients accordingly.

    Effective Fock matrix is built with one-particle density matrix (see
    also :func:`mcscf.casci.get_fock`). For state-average CASCI/CASSCF object,
    the canonicalized orbitals are based on the state-average density matrix.
    To obtain canonicalized orbitals for an individual state, you need to pass
    "casdm1" of the specific state to this function.

    Args:
        mc: a CASSCF/CASCI object or RHF object

    Kwargs:
        mo_coeff (ndarray): orbitals that span the core, active and external
            space.
        ci (ndarray): CI coefficients (or objects to represent the CI
            wavefunctions in DMRG/QMC-MCSCF calculations).
        eris: Integrals for the MCSCF object. Input this object to reduce the
            overhead of computing integrals. It can be generated by
            :func:`mc.ao2mo` method.
        sort (bool): Whether the canonicalized orbitals are sorted based on
            the orbital energy (diagonal part of the effective Fock matrix)
            within each subspace (core, active, external). If point group
            symmetry is not available in the system, orbitals are always
            sorted. When point group symmetry is available, sort=False will
            preserve the symmetry label of input orbitals and only sort the
            orbitals in each symmetry sector. sort=True will reorder all
            orbitals over all symmetry sectors in each subspace and the
            symmetry labels may be changed.
        cas_natorb (bool): Whether to transform active orbitals to natural
            orbitals. If enabled, the output orbitals in active space are
            transformed to natural orbitals and CI coefficients are updated
            accordingly.
        casdm1 (ndarray): 1-particle density matrix in active space. This
            density matrix is used to build effective fock matrix. Without
            input casdm1, the density matrix is computed with the input ci
            coefficients/object. If neither ci nor casdm1 were given, density
            matrix is computed by :func:`mc.fcisolver.make_rdm1` method. For
            state-average CASCI/CASCF calculation, this results in a set of
            canonicalized orbitals of state-average effective Fock matrix.
            To canonicalize the orbitals for one particular state, you can
            assign the density matrix of that state to the kwarg casdm1.
        stav_dm1 (bool): Use state-average 1-particle density matrix for
            computing Fock matrices and natural orbitals

    Returns:
        A tuple, (natural orbitals, CI coefficients, orbital energies)
        The orbital energies are the diagonal terms of effective Fock matrix.
    '''
    from pyscf.mcscf import addons
    log = logger.new_logger(mc, verbose)

    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if casdm1 is None:
        if (isinstance(ci, (list, tuple, RANGE_TYPE)) and
                not isinstance(mc.fcisolver, addons.StateAverageFCISolver)):
            if stav_dm1:
                log.warn('Mulitple states found in CASCI solver. '
                         'Use state-average 1RDM  to compute the Fock matrix'
                         ' and natural orbitals in the active space.')
                casdm1 = mc.fcisolver.make_rdm1(ci[0], mc.ncas, mc.nelecas)
                for root in range(1, len(ci)):
                    casdm1 += mc.fcisolver.make_rdm1(ci[root], mc.ncas,
                                                     mc.nelecas)
                casdm1 /= len(ci)
            else:
                log.warn('Mulitple states found in CASCI solver. '
                         'First state is used to compute the Fock matrix'
                         ' and natural orbitals in active space.')
                casdm1 = mc.fcisolver.make_rdm1(ci[0], mc.ncas, mc.nelecas)
        else:
            casdm1 = mc.fcisolver.make_rdm1(ci, mc.ncas, mc.nelecas)

    ncore = mc.ncore
    nocc = ncore + mc.ncas
    nmo = mo_coeff.shape[1]
    fock_ao = mc.get_fock(mo_coeff, ci, eris, casdm1, verbose)

    if cas_natorb:
        mo_coeff1, ci, mc.mo_occ = mc.cas_natorb(mo_coeff, ci, eris, sort, casdm1,
                                                 verbose, with_meta_lowdin)
    else:
        # Keep the active space unchanged by default.  The rotation in active space
        # may cause problem for external CI solver eg DMRG.
        mo_coeff1 = mo_coeff.copy()
        log.info('Density matrix diagonal elements %s', casdm1.diagonal())

    mo_energy = numpy.einsum('pi,pi->i', mo_coeff1.conj(), fock_ao.dot(mo_coeff1))

    if getattr(mo_coeff, 'orbsym', None) is not None:
        orbsym = mo_coeff.orbsym
    else:
        orbsym = numpy.zeros(nmo, dtype=int)
    extrasym = getattr(mc, 'extrasym', None)
    if extrasym is not None:
        orbsym_extra = numpy.asarray([str(i1) + str(i2)
                                      for i1, i2 in zip(orbsym, extrasym)])
    else:
        orbsym_extra = orbsym

    def _diag_subfock_(idx):
        if idx.size > 1:
            c = mo_coeff1[:,idx]
            fock = reduce(numpy.dot, (c.conj().T, fock_ao, c))
            # note the last argument orbysm is needed by mc1step_symm._eig
            w, c = mc._eig(fock, None, None, orbsym_extra[idx])

            if sort:
                sub_order = numpy.argsort(w.round(9), kind='mergesort')
                w = w[sub_order]
                c = c[:,sub_order]
                orbsym[idx] = orbsym[idx][sub_order]

            mo_coeff1[:,idx] = mo_coeff1[:,idx].dot(c)
            mo_energy[idx] = w

    mask = numpy.ones(nmo, dtype=bool)
    frozen = getattr(mc, 'frozen', None)
    if frozen is not None:
        if isinstance(frozen, (int, numpy.integer)):
            mask[:frozen] = False
        else:
            mask[frozen] = False
    core_idx = numpy.where(mask[:ncore])[0]
    vir_idx = numpy.where(mask[nocc:])[0] + nocc
    _diag_subfock_(core_idx)
    _diag_subfock_(vir_idx)

    # orbsym is required only for symmetry-adapted methods. Here to use
    # mo_coeff.orbsym to test if a symmetry-adapted calculation.
    if getattr(mo_coeff, 'orbsym', None) is not None:
        mo_coeff1 = lib.tag_array(mo_coeff1, orbsym=orbsym)

    if log.verbose >= logger.DEBUG:
        for i in range(nmo):
            log.debug('i = %d  <i|F|i> = %12.8f', i+1, mo_energy[i])
# still return ci coefficients, in case the canonicalization function changed
# cas orbitals, the ci coefficients should also be updated.
    return mo_coeff1, ci, mo_energy


def kernel(casci, mo_coeff=None, ci0=None, verbose=logger.NOTE, envs=None):
    '''CASCI solver

    Args:
        casci: CASCI or CASSCF object

        mo_coeff : ndarray
            orbitals to construct active space Hamiltonian
        ci0 : ndarray or custom types
            FCI sovler initial guess. For external FCI-like solvers, it can be
            overloaded different data type. For example, in the state-average
            FCI solver, ci0 is a list of ndarray. In other solvers such as
            DMRGCI solver, SHCI solver, ci0 are custom types.

    kwargs:
        envs: dict
            The variable envs is created (for PR 807) to passes MCSCF runtime
            environment variables to SHCI solver. For solvers which do not
            need this parameter, a kwargs should be created in kernel method
            and "envs" pop in kernel function
    '''
    if mo_coeff is None: mo_coeff = casci.mo_coeff
    if ci0 is None: ci0 = casci.ci

    log = logger.new_logger(casci, verbose)
    t0 = (logger.process_clock(), logger.perf_counter())
    log.debug('Start CASCI')

    ncas = casci.ncas
    nelecas = casci.nelecas

    # 2e
    eri_cas = casci.get_h2eff(mo_coeff)
    t1 = log.timer('integral transformation to CAS space', *t0)

    # 1e
    h1eff, energy_core = casci.get_h1eff(mo_coeff)
    log.debug('core energy = %.15g', energy_core)
    t1 = log.timer('effective h1e in CAS space', *t1)

    if h1eff.shape[0] != ncas:
        raise RuntimeError('Active space size error. nmo=%d ncore=%d ncas=%d' %
                           (mo_coeff.shape[1], casci.ncore, ncas))

    # FCI
    max_memory = max(400, casci.max_memory-lib.current_memory()[0])
    e_tot, fcivec = casci.fcisolver.kernel(h1eff, eri_cas, ncas, nelecas,
                                           ci0=ci0, verbose=log,
                                           max_memory=max_memory,
                                           ecore=energy_core)

    t1 = log.timer('FCI solver', *t1)
    e_cas = e_tot - energy_core
    return e_tot, e_cas, fcivec


def as_scanner(mc):
    '''Generating a scanner for CASCI PES.

    The returned solver is a function. This function requires one argument
    "mol" as input and returns total CASCI energy.

    The solver will automatically use the results of last calculation as the
    initial guess of the new calculation.  All parameters of MCSCF object
    are automatically applied in the solver.

    Note scanner has side effects.  It may change many underlying objects
    (_scf, with_df, with_x2c, ...) during calculation.

    Examples:

    >>> from pyscf import gto, scf, mcscf
    >>> mf = scf.RHF(gto.Mole().set(verbose=0))
    >>> mc_scanner = mcscf.CASCI(mf, 4, 4).as_scanner()
    >>> mc_scanner(gto.M(atom='N 0 0 0; N 0 0 1.1'))
    >>> mc_scanner(gto.M(atom='N 0 0 0; N 0 0 1.5'))
    '''
    if isinstance(mc, lib.SinglePointScanner):
        return mc

    logger.info(mc, 'Create scanner for %s', mc.__class__)
    name = mc.__class__.__name__ + CASCI_Scanner.__name_mixin__
    return lib.set_class(CASCI_Scanner(mc), (CASCI_Scanner, mc.__class__), name)

class CASCI_Scanner(lib.SinglePointScanner):
    def __init__(self, mc):
        self.__dict__.update(mc.__dict__)
        self._scf = mc._scf.as_scanner()

    def __call__(self, mol_or_geom, mo_coeff=None, ci0=None):
        if isinstance(mol_or_geom, gto.MoleBase):
            mol = mol_or_geom
        else:
            mol = self.mol.set_geom_(mol_or_geom, inplace=False)

        self.reset(mol)
        for key in ('with_df', 'with_x2c', 'with_solvent', 'with_dftd3'):
            sub_mod = getattr(self, key, None)
            if sub_mod:
                sub_mod.reset(mol)

        if mo_coeff is None:
            mf_scanner = self._scf
            mf_scanner(mol)
            mo_coeff = mf_scanner.mo_coeff
        if ci0 is None:
            ci0 = self.ci
        self.mol = mol
        e_tot = self.kernel(mo_coeff, ci0)[0]
        return e_tot


class CASBase(lib.StreamObject):
    '''CASCI/CASSCF

    Args:
        mf_or_mol : SCF object or Mole object
            SCF or Mole to define the problem size.
        ncas : int
            Number of active orbitals.
        nelecas : int or a pair of int
            Number of electrons in active space.

    Kwargs:
        ncore : int
            Number of doubly occupied core orbitals. If not presented, this
            parameter can be automatically determined.

    Attributes:
        verbose : int
            Print level.  Default value equals to :class:`Mole.verbose`.
        max_memory : float or int
            Allowed memory in MB.  Default value equals to :class:`Mole.max_memory`.
        ncas : int
            Active space size.
        nelecas : tuple of int
            Active (nelec_alpha, nelec_beta)
        ncore : int or tuple of int
            Core electron number.  In UHF-CASSCF, it's a tuple to indicate the different core electron numbers.
        natorb : bool
            Whether to transform natural orbitals in active space.
            Note: when CASCI/CASSCF are combined with DMRG solver or selected
            CI solver, enabling this parameter may slightly change the total energy.
            False by default.
        canonicalization : bool
            Whether to canonicalize orbitals in core and external space
            against the general Fock matrix.
            The orbitals in active space are NOT transformed by default. To
            get the natural orbitals in active space, the attribute .natorb
            needs to be enabled.
            True by default.
        sorting_mo_energy : bool
            Whether to sort the orbitals based on the diagonal elements of the
            general Fock matrix.  Default is False.
        fcisolver : an instance of :class:`FCISolver`
            The pyscf.fci module provides several FCISolver for different scenario.  Generally,
            fci.direct_spin1.FCISolver can be used for all RHF-CASSCF.  However, a proper FCISolver
            can provide better performance and better numerical stability.  One can either use
            :func:`fci.solver` function to pick the FCISolver by the program or manually assigen
            the FCISolver to this attribute, e.g.

            >>> from pyscf import fci
            >>> mc = mcscf.CASSCF(mf, 4, 4)
            >>> mc.fcisolver = fci.solver(mol, singlet=True)
            >>> mc.fcisolver = fci.direct_spin1.FCISolver(mol)

            You can control FCISolver by setting e.g.::

                >>> mc.fcisolver.max_cycle = 30
                >>> mc.fcisolver.conv_tol = 1e-7

            For more details of the parameter for FCISolver, See :mod:`fci`.

    Saved results

        e_tot : float
            Total MCSCF energy (electronic energy plus nuclear repulsion)
        e_cas : float
            CAS space FCI energy
        ci : ndarray
            CAS space FCI coefficients
        mo_coeff : ndarray
            When canonicalization is specified, the orbitals are canonical
            orbitals which make the general Fock matrix (Fock operator on top
            of MCSCF 1-particle density matrix) diagonalized within each
            subspace (core, active, external).  If natorb (natural orbitals in
            active space) is specified, the active segment of the mo_coeff is
            natural orbitals.
        mo_energy : ndarray
            Diagonal elements of general Fock matrix (in mo_coeff
            representation).
        mo_occ : ndarray
            Occupation numbers of natural orbitals if natorb is specified.

    Examples:

    >>> from pyscf import gto, scf, mcscf
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1', basis='ccpvdz', verbose=0)
    >>> mf = scf.RHF(mol)
    >>> mf.scf()
    >>> mc = mcscf.CASCI(mf, 6, 6)
    >>> mc.kernel()[0]
    -108.980200816243354
    '''

    natorb = getattr(__config__, 'mcscf_casci_CASCI_natorb', False)
    canonicalization = getattr(__config__, 'mcscf_casci_CASCI_canonicalization', True)
    sorting_mo_energy = getattr(__config__, 'mcscf_casci_CASCI_sorting_mo_energy', False)

    _keys = {
        'natorb', 'canonicalization', 'sorting_mo_energy', 'mol', 'max_memory',
        'ncas', 'nelecas', 'ncore', 'fcisolver', 'frozen', 'extrasym',
        'e_tot', 'e_cas', 'ci', 'mo_coeff', 'mo_energy', 'mo_occ', 'converged',
    }

    def __init__(self, mf_or_mol, ncas=0, nelecas=0, ncore=None):
        if isinstance(mf_or_mol, gto.Mole):
            mf = scf.RHF(mf_or_mol)
        else:
            mf = mf_or_mol

        mol = mf.mol
        self.mol = mol
        self._scf = mf
        self.verbose = mol.verbose
        self.stdout = mol.stdout
        self.max_memory = mf.max_memory
        self.ncas = ncas
        if isinstance(nelecas, (int, numpy.integer)):
            nelecb = (nelecas-mol.spin)//2
            neleca = nelecas - nelecb
            self.nelecas = (neleca, nelecb)
        else:
            self.nelecas = (nelecas[0],nelecas[1])
        self.ncore = ncore
        singlet = (getattr(__config__, 'mcscf_casci_CASCI_fcisolver_direct_spin0', False)
                   and self.nelecas[0] == self.nelecas[1])  # leads to direct_spin1
        self.fcisolver = fci.solver(mol, singlet, symm=False)
# CI solver parameters are set in fcisolver object
        self.fcisolver.lindep = getattr(__config__,
                                        'mcscf_casci_CASCI_fcisolver_lindep', 1e-12)
        self.fcisolver.max_cycle = getattr(__config__,
                                           'mcscf_casci_CASCI_fcisolver_max_cycle', 200)
        self.fcisolver.conv_tol = getattr(__config__,
                                          'mcscf_casci_CASCI_fcisolver_conv_tol', 1e-8)
        self.frozen = None
        self.extrasym = None

##################################################
# don't modify the following attributes, they are not input options
        self.e_tot = 0
        self.e_cas = None
        self.ci = None
        self.mo_coeff = mf.mo_coeff
        self.mo_energy = mf.mo_energy
        self.mo_occ = None
        self.converged = False

    @property
    def ncore(self):
        if self._ncore is None:
            ncorelec = self.mol.nelectron - sum(self.nelecas)
            assert ncorelec % 2 == 0
            assert ncorelec >= 0
            return ncorelec // 2
        else:
            return self._ncore
    @ncore.setter
    def ncore(self, x):
        assert x is None or isinstance(x, (int, numpy.integer))
        assert x is None or x >= 0
        self._ncore = x

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('')
        log.info('******** %s ********', self.__class__)
        ncore = self.ncore
        ncas = self.ncas
        nvir = self.mo_coeff.shape[1] - ncore - ncas
        log.info('CAS (%de+%de, %do), ncore = %d, nvir = %d',
                 self.nelecas[0], self.nelecas[1], ncas, ncore, nvir)
        if self.frozen is not None:
            log.info('frozen orbitals %s', str(self.frozen))
        if self.extrasym is not None:
            log.info('extra symmetry labels:\n%s', str(self.extrasym))
        log.info('natorb = %s', self.natorb)
        log.info('canonicalization = %s', self.canonicalization)
        log.info('sorting_mo_energy = %s', self.sorting_mo_energy)
        log.info('max_memory %d MB (current use %d MB)',
                 self.max_memory, lib.current_memory()[0])
        if getattr(self.fcisolver, 'dump_flags', None):
            self.fcisolver.dump_flags(log.verbose)
        if self.mo_coeff is None:
            log.error('Orbitals for CASCI are not specified. The relevant SCF '
                      'object may not be initialized.')

        if (getattr(self._scf, 'with_solvent', None) and
            not getattr(self, 'with_solvent', None)):
            log.warn('''Solvent model %s was found at SCF level but not applied to the CASCI object.
The SCF solvent model will not be applied to the current CASCI calculation.
To enable the solvent model for CASCI, the following code needs to be called
        from pyscf import solvent
        mc = mcscf.CASCI(...)
        mc = solvent.ddCOSMO(mc)
''',
                     self._scf.with_solvent.__class__)
        return self

    def check_sanity(self):
        super().check_sanity()
        assert self.ncas > 0
        ncore = self.ncore
        nvir = self.mo_coeff.shape[1] - ncore - self.ncas
        assert ncore >= 0
        assert nvir >= 0
        assert ncore * 2 + sum(self.nelecas) == self.mol.nelectron
        assert 0 <= self.nelecas[0] <= self.ncas
        assert 0 <= self.nelecas[1] <= self.ncas
        return self

    def reset(self, mol=None):
        if mol is not None:
            self.mol = mol
            self.fcisolver.mol = mol
        self._scf.reset(mol)
        return self

    def energy_nuc(self):
        return self._scf.energy_nuc()

    def get_hcore(self, mol=None):
        return self._scf.get_hcore(mol)

    @lib.with_doc(scf.hf.get_jk.__doc__)
    def get_jk(self, mol, dm, hermi=1, with_j=True, with_k=True, omega=None):
        return self._scf.get_jk(mol, dm, hermi,
                                with_j=with_j, with_k=with_k, omega=omega)

    @lib.with_doc(scf.hf.get_veff.__doc__)
    def get_veff(self, mol=None, dm=None, hermi=1):
        if mol is None: mol = self.mol
        if dm is None:
            mocore = self.mo_coeff[:,:self.ncore]
            dm = numpy.dot(mocore, mocore.conj().T) * 2
# don't call self._scf.get_veff because _scf might be DFT object
        vj, vk = self.get_jk(mol, dm, hermi)
        return vj - vk * .5

    def _eig(self, h, *args):
        return scf.hf.eig(h, None)

    def get_h2cas(self, mo_coeff=None):
        '''An alias of get_h2eff method'''
        return self.get_h2eff(mo_coeff)

    def get_h2eff(self, mo_coeff=None):
        '''Compute the active space two-particle Hamiltonian.
        '''
        raise NotImplementedError

    def ao2mo(self, mo_coeff=None):
        '''Compute the active space two-particle Hamiltonian.
        '''
        raise NotImplementedError

    def get_h1cas(self, mo_coeff=None, ncas=None, ncore=None):
        '''An alias of get_h1eff method'''
        return self.get_h1eff(mo_coeff, ncas, ncore)

    get_h1eff = h1e_for_cas = h1e_for_cas

    def casci(self, mo_coeff=None, ci0=None, verbose=None):
        raise NotImplementedError

    def kernel(self, mo_coeff=None, ci0=None, verbose=None):
        '''
        Returns:
            Five elements, they are
            total energy,
            active space CI energy,
            the active space FCI wavefunction coefficients or DMRG wavefunction ID,
            the MCSCF canonical orbital coefficients,
            the MCSCF canonical orbital energies (diagonal elements of general Fock matrix).

        They are attributes of mcscf object, which can be accessed by
        .e_tot, .e_cas, .ci, .mo_coeff, .mo_energy
        '''
        raise NotImplementedError

    def _finalize(self):
        log = logger.Logger(self.stdout, self.verbose)
        if log.verbose >= logger.NOTE and getattr(self.fcisolver, 'spin_square', None):
            if isinstance(self.e_cas, (float, numpy.number)):
                try:
                    ss = self.fcisolver.spin_square(self.ci, self.ncas, self.nelecas)
                    log.note('CASCI E = %#.15g  E(CI) = %#.15g  S^2 = %.7f',
                             self.e_tot, self.e_cas, ss[0])
                except NotImplementedError:
                    log.note('CASCI E = %#.15g  E(CI) = %#.15g',
                             self.e_tot, self.e_cas)
            else:
                for i, e in enumerate(self.e_cas):
                    try:
                        ss = self.fcisolver.spin_square(self.ci[i], self.ncas, self.nelecas)
                        log.note('CASCI state %3d  E = %#.15g  E(CI) = %#.15g  S^2 = %.7f',
                                 i, self.e_tot[i], e, ss[0])
                    except NotImplementedError:
                        log.note('CASCI state %3d  E = %#.15g  E(CI) = %#.15g',
                                 i, self.e_tot[i], e)

        else:
            if isinstance(self.e_cas, (float, numpy.number)):
                log.note('CASCI E = %#.15g  E(CI) = %#.15g', self.e_tot, self.e_cas)
            else:
                for i, e in enumerate(self.e_cas):
                    log.note('CASCI state %3d  E = %#.15g  E(CI) = %#.15g',
                             i, self.e_tot[i], e)
        return self

    @lib.with_doc(cas_natorb.__doc__)
    def cas_natorb(self, mo_coeff=None, ci=None, eris=None, sort=False,
                   casdm1=None, verbose=None, with_meta_lowdin=WITH_META_LOWDIN):
        return cas_natorb(self, mo_coeff, ci, eris, sort, casdm1, verbose,
                          with_meta_lowdin)
    @lib.with_doc(cas_natorb.__doc__)
    def cas_natorb_(self, mo_coeff=None, ci=None, eris=None, sort=False,
                    casdm1=None, verbose=None, with_meta_lowdin=WITH_META_LOWDIN):
        self.mo_coeff, self.ci, self.mo_occ = cas_natorb(self, mo_coeff, ci, eris,
                                                         sort, casdm1, verbose)
        return self.mo_coeff, self.ci, self.mo_occ

    def get_fock(self, mo_coeff=None, ci=None, eris=None, casdm1=None,
                 verbose=None):
        return get_fock(self, mo_coeff, ci, eris, casdm1, verbose)

    canonicalize = canonicalize

    @lib.with_doc(canonicalize.__doc__)
    def canonicalize_(self, mo_coeff=None, ci=None, eris=None, sort=False,
                      cas_natorb=False, casdm1=None, verbose=None,
                      with_meta_lowdin=WITH_META_LOWDIN):
        self.mo_coeff, ci, self.mo_energy = \
                canonicalize(self, mo_coeff, ci, eris,
                             sort, cas_natorb, casdm1, verbose, with_meta_lowdin)
        if cas_natorb:  # When active space is changed, the ci solution needs to be updated
            self.ci = ci
        return self.mo_coeff, ci, self.mo_energy

    analyze = analyze

    @lib.with_doc(addons.sort_mo.__doc__)
    def sort_mo(self, caslst, mo_coeff=None, base=1):
        if mo_coeff is None: mo_coeff = self.mo_coeff
        return addons.sort_mo(self, mo_coeff, caslst, base)

    @lib.with_doc(addons.state_average.__doc__)
    def state_average_(self, weights=(0.5,0.5), wfnsym=None):
        addons.state_average_(self, weights, wfnsym)
        return self
    @lib.with_doc(addons.state_average.__doc__)
    def state_average(self, weights=(0.5,0.5), wfnsym=None):
        return addons.state_average(self, weights, wfnsym)

    @lib.with_doc(addons.state_specific_.__doc__)
    def state_specific_(self, state=1):
        addons.state_specific(self, state)
        return self

    def make_rdm1s(self, mo_coeff=None, ci=None, ncas=None, nelecas=None,
                   ncore=None, **kwargs):
        '''One-particle density matrices for alpha and beta spin on AO basis
        '''
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if ci is None: ci = self.ci
        if ncas is None: ncas = self.ncas
        if nelecas is None: nelecas = self.nelecas
        if ncore is None: ncore = self.ncore

        casdm1a, casdm1b = self.fcisolver.make_rdm1s(ci, ncas, nelecas)
        mocore = mo_coeff[:,:ncore]
        mocas = mo_coeff[:,ncore:ncore+ncas]
        dm1b = numpy.dot(mocore, mocore.conj().T)
        dm1a = dm1b + reduce(numpy.dot, (mocas, casdm1a, mocas.conj().T))
        dm1b += reduce(numpy.dot, (mocas, casdm1b, mocas.conj().T))
        return dm1a, dm1b

    def make_rdm1(self, mo_coeff=None, ci=None, ncas=None, nelecas=None,
                  ncore=None, **kwargs):
        '''One-particle density matrix in AO representation
        '''
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if ci is None: ci = self.ci
        if ncas is None: ncas = self.ncas
        if nelecas is None: nelecas = self.nelecas
        if ncore is None: ncore = self.ncore

        casdm1 = self.fcisolver.make_rdm1(ci, ncas, nelecas)
        mocore = mo_coeff[:,:ncore]
        mocas = mo_coeff[:,ncore:ncore+ncas]
        dm1 = numpy.dot(mocore, mocore.conj().T) * 2
        dm1 = dm1 + reduce(numpy.dot, (mocas, casdm1, mocas.conj().T))
        return dm1

    def fix_spin_(self, shift=PENALTY, ss=None):
        r'''Use level shift to control FCI solver spin.

        .. math::

            (H + shift*S^2) |\Psi\rangle = E |\Psi\rangle

        Kwargs:
            shift : float
                Energy penalty for states which have wrong spin
            ss : number
                S^2 expection value == s*(s+1)
        '''
        fci.addons.fix_spin_(self.fcisolver, shift, ss)
        return self
    fix_spin = fix_spin_

    def density_fit(self, auxbasis=None, with_df=None):
        from pyscf.mcscf import df
        return df.density_fit(self, auxbasis, with_df)

    def sfx2c1e(self):
        from pyscf.x2c import sfx2c1e
        self._scf = sfx2c1e.sfx2c1e(self._scf).run()
        self.mo_coeff = self._scf.mo_coeff
        self.mo_energy = self._scf.mo_energy
        return self
    x2c = x2c1e = sfx2c1e

    def nuc_grad_method(self):
        raise NotImplementedError

class CASCI(CASBase):
    def get_h2eff(self, mo_coeff=None):
        '''Compute the active space two-particle Hamiltonian.
        '''
        ncore = self.ncore
        ncas = self.ncas
        nocc = ncore + ncas
        if mo_coeff is None:
            ncore = self.ncore
            mo_coeff = self.mo_coeff[:,ncore:nocc]
        elif mo_coeff.shape[1] != ncas:
            mo_coeff = mo_coeff[:,ncore:nocc]

        if hasattr(self._scf, '_eri') and self._scf._eri is not None:
            eri = ao2mo.full(self._scf._eri, mo_coeff,
                             max_memory=self.max_memory)
        else:
            eri = ao2mo.full(self.mol, mo_coeff, verbose=self.verbose,
                             max_memory=self.max_memory)
        return eri

    def casci(self, mo_coeff=None, ci0=None, verbose=None):
        return self.kernel(mo_coeff, ci0, verbose)
    def kernel(self, mo_coeff=None, ci0=None, verbose=None):
        '''
        Returns:
            Five elements, they are
            total energy,
            active space CI energy,
            the active space FCI wavefunction coefficients or DMRG wavefunction ID,
            the MCSCF canonical orbital coefficients,
            the MCSCF canonical orbital energies (diagonal elements of general Fock matrix).

        They are attributes of mcscf object, which can be accessed by
        .e_tot, .e_cas, .ci, .mo_coeff, .mo_energy
        '''
        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        else:
            self.mo_coeff = mo_coeff
        if ci0 is None:
            ci0 = self.ci
        log = logger.new_logger(self, verbose)

        self.check_sanity()
        self.dump_flags(log)

        self.e_tot, self.e_cas, self.ci = \
                kernel(self, mo_coeff, ci0=ci0, verbose=log)

        if self.canonicalization:
            self.canonicalize_(mo_coeff, self.ci,
                               sort=self.sorting_mo_energy,
                               cas_natorb=self.natorb, verbose=log)
        elif self.natorb:
            # FIXME (pyscf-2.0): Whether to transform natural orbitals in
            # active space when this flag is enabled?
            log.warn('The attribute .natorb of mcscf object affects only the '
                     'orbital canonicalization.\n'
                     'If you would like to get natural orbitals in active space '
                     'without touching core and external orbitals, an explicit '
                     'call to mc.cas_natorb_() is required')

        if getattr(self.fcisolver, 'converged', None) is not None:
            self.converged = numpy.all(self.fcisolver.converged)
            if self.converged:
                log.info('CASCI converged')
            else:
                log.info('CASCI not converged')
        else:
            self.converged = True
        self._finalize()
        return self.e_tot, self.e_cas, self.ci, self.mo_coeff, self.mo_energy

    as_scanner = as_scanner

    def nuc_grad_method(self):
        from pyscf.grad import casci
        return casci.Gradients(self)

    to_gpu = lib.to_gpu

scf.hf.RHF.CASCI = scf.rohf.ROHF.CASCI = lib.class_as_method(CASCI)
scf.uhf.UHF.CASCI = None

del (WITH_META_LOWDIN, LARGE_CI_TOL, PENALTY)



# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Periodic G0W0 approximation
'''

from pyscf.pbc.gw import krgw_ac
from pyscf.pbc.gw import kugw_ac
from pyscf.pbc.gw import krgw_cd
from pyscf.pbc import scf

def KRGW(mf, freq_int='ac', frozen=None):
    # mf = mf.to_rhf()
    if freq_int.lower() == 'ac':
        return krgw_ac.KRGWAC(mf, frozen)
    elif freq_int.lower() == 'cd':
        return krgw_cd.KRGWCD(mf, frozen)
    else:
        raise RuntimeError("GW frequency integration method %s not recognized. "
                           "With PBC, options are 'ac' and 'cd'."%(freq_int))

def KUGW(mf, freq_int='ac', frozen=None):
    # mf = mf.to_uhf()
    if freq_int.lower() == 'ac':
        return kugw_ac.KUGWAC(mf, frozen)
    elif freq_int.lower() == 'cd':
        raise RuntimeError('GWCD does not support UHF or UKS methods.')
    else:
        raise RuntimeError("GW frequency integration method %s not recognized. "
                           "With PBC, options are 'ac' and 'cd'."%(freq_int))

KGW = KRGW



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import ctypes
import numpy as np
import scipy
import scipy.linalg
from pyscf import lib
from pyscf.lib import logger
from pyscf.gto import ATM_SLOTS, BAS_SLOTS, ATOM_OF, PTR_COORD
from pyscf.pbc.lib.kpts_helper import get_kconserv, get_kconserv3  # noqa
from pyscf.pbc.lib.kpts_helper import intersection
from pyscf import __config__

FFT_ENGINE = getattr(__config__, 'pbc_tools_pbc_fft_engine', 'NUMPY+BLAS')

def _fftn_blas(f, mesh):
    assert f.ndim == 4
    mx, my, mz = mesh
    expRGx = np.exp(-2j*np.pi*np.arange(mx)[:,None] * np.fft.fftfreq(mx))
    expRGy = np.exp(-2j*np.pi*np.arange(my)[:,None] * np.fft.fftfreq(my))
    expRGz = np.exp(-2j*np.pi*np.arange(mz)[:,None] * np.fft.fftfreq(mz))
    blksize = max(int(1e5 / (mx * my * mz)), 8) * 4
    n = f.shape[0]
    out = np.empty((n, mx*my*mz), dtype=np.complex128)
    buf = np.empty((blksize, mx*my*mz), dtype=np.complex128)
    for i0, i1 in lib.prange(0, n, blksize):
        ni = i1 - i0
        buf1 = buf[:ni]
        out1 = out[i0:i1]
        g = lib.transpose(f[i0:i1].reshape(ni,-1), out=buf1.reshape(-1,ni))
        g = lib.dot(g.reshape(mx,-1).T, expRGx, c=out1.reshape(-1,mx))
        g = lib.dot(g.reshape(my,-1).T, expRGy, c=buf1.reshape(-1,my))
        g = lib.dot(g.reshape(mz,-1).T, expRGz, c=out1.reshape(-1,mz))
    return out.reshape(n, *mesh)

def _ifftn_blas(g, mesh):
    assert g.ndim == 4
    mx, my, mz = mesh
    expRGx = np.exp(2j*np.pi*np.fft.fftfreq(mx)[:,None] * np.arange(mx))
    expRGy = np.exp(2j*np.pi*np.fft.fftfreq(my)[:,None] * np.arange(my))
    expRGz = np.exp(2j*np.pi*np.fft.fftfreq(mz)[:,None] * np.arange(mz))
    blksize = max(int(1e5 / (mx * my * mz)), 8) * 4
    n = g.shape[0]
    out = np.empty((n, mx*my*mz), dtype=np.complex128)
    buf = np.empty((blksize, mx*my*mz), dtype=np.complex128)
    for i0, i1 in lib.prange(0, n, blksize):
        ni = i1 - i0
        buf1 = buf[:ni]
        out1 = out[i0:i1]
        f = lib.transpose(g[i0:i1].reshape(ni,-1), out=buf1.reshape(-1,ni))
        f = lib.dot(f.reshape(mx,-1).T, expRGx, 1./mx, c=out1.reshape(-1,mx))
        f = lib.dot(f.reshape(my,-1).T, expRGy, 1./my, c=buf1.reshape(-1,my))
        f = lib.dot(f.reshape(mz,-1).T, expRGz, 1./mz, c=out1.reshape(-1,mz))
    return out.reshape(n, *mesh)

nproc = lib.num_threads()

def _fftn_wrapper(a):  # noqa
    return scipy.fft.fftn(a, axes=(1,2,3), workers=nproc)

def _ifftn_wrapper(a):  # noqa
    return scipy.fft.ifftn(a, axes=(1,2,3), workers=nproc)

if FFT_ENGINE == 'FFTW':
    try:
        libfft = lib.load_library('libfft')
    except OSError:
        raise RuntimeError("Failed to load libfft")

    def _copy_d2z(a):
        fn = libfft._copy_d2z
        out = np.empty(a.shape, dtype=np.complex128)
        fn(out.ctypes.data_as(ctypes.c_void_p),
           a.ctypes.data_as(ctypes.c_void_p),
           ctypes.c_size_t(a.size))
        return out

    def _complex_fftn_fftw(f, mesh, func):
        if f.dtype == np.double and f.flags.c_contiguous:
            # np.asarray or np.astype is too slow
            f = _copy_d2z(f)
        else:
            f = np.asarray(f, order='C', dtype=np.complex128)
        mesh = np.asarray(mesh, order='C', dtype=np.int32)
        rank = len(mesh)
        out = np.empty_like(f)
        fn = getattr(libfft, func)
        for i, fi in enumerate(f):
            fn(fi.ctypes.data_as(ctypes.c_void_p),
               out[i].ctypes.data_as(ctypes.c_void_p),
               mesh.ctypes.data_as(ctypes.c_void_p),
               ctypes.c_int(rank))
        return out

    def _fftn_wrapper(a):  # noqa
        mesh = a.shape[1:]
        return _complex_fftn_fftw(a, mesh, 'fft')
    def _ifftn_wrapper(a):  # noqa
        mesh = a.shape[1:]
        return _complex_fftn_fftw(a, mesh, 'ifft')

elif FFT_ENGINE == 'PYFFTW':
    # Note: pyfftw is likely slower than scipy.fft in multi-threading environments
    try:
        import pyfftw
        pyfftw.config.PLANNER_EFFORT = 'FFTW_MEASURE'
        pyfftw.interfaces.cache.enable()
        def _fftn_wrapper(a):  # noqa
            return pyfftw.interfaces.numpy_fft.fftn(a, axes=(1,2,3), threads=nproc)
        def _ifftn_wrapper(a):  # noqa
            return pyfftw.interfaces.numpy_fft.ifftn(a, axes=(1,2,3), threads=nproc)
    except ImportError:
        print('PyFFTW not installed. SciPy fft module will be used.')

elif FFT_ENGINE == 'NUMPY+BLAS':
    _EXCLUDE = [17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79,
                83, 89, 97,101,103,107,109,113,127,131,137,139,149,151,157,163,
                167,173,179,181,191,193,197,199,211,223,227,229,233,239,241,251,
                257,263,269,271,277,281,283,293]
    _EXCLUDE = set(_EXCLUDE + [n*2 for n in _EXCLUDE[:30]] + [n*3 for n in _EXCLUDE[:20]])
    def _fftn_wrapper(a):  # noqa
        mesh = a.shape[1:]
        if mesh[0] in _EXCLUDE and mesh[1] in _EXCLUDE and mesh[2] in _EXCLUDE:
            return _fftn_blas(a, mesh)
        else:
            return scipy.fft.fftn(a, axes=(1,2,3), workers=nproc)
    def _ifftn_wrapper(a):  # noqa
        mesh = a.shape[1:]
        if mesh[0] in _EXCLUDE and mesh[1] in _EXCLUDE and mesh[2] in _EXCLUDE:
            return _ifftn_blas(a, mesh)
        else:
            return scipy.fft.ifftn(a, axes=(1,2,3), workers=nproc)

elif FFT_ENGINE == 'BLAS':
    def _fftn_wrapper(a):  # noqa
        mesh = a.shape[1:]
        return _fftn_blas(a, mesh)
    def _ifftn_wrapper(a):  # noqa
        mesh = a.shape[1:]
        return _ifftn_blas(a, mesh)


def fft(f, mesh):
    '''Perform the 3D FFT from real (R) to reciprocal (G) space.

    After FFT, (u, v, w) -> (j, k, l).
    (jkl) is in the index order of Gv.

    FFT normalization factor is 1., as in MH and in `numpy.fft`.

    Args:
        f : (nx*ny*nz,) ndarray
            The function to be FFT'd, flattened to a 1D array corresponding
            to the index order of :func:`cartesian_prod`.
        mesh : (3,) ndarray of ints (= nx,ny,nz)
            The number G-vectors along each direction.

    Returns:
        (nx*ny*nz,) ndarray
            The FFT 1D array in same index order as Gv (natural order of
            numpy.fft).

    '''
    if f.size == 0:
        return np.zeros_like(f)

    f3d = f.reshape(-1, *mesh)
    assert (f3d.shape[0] == 1 or f[0].size == f3d[0].size)
    g3d = _fftn_wrapper(f3d)
    ngrids = np.prod(mesh)
    if f.ndim == 1 or (f.ndim == 3 and f.size == ngrids):
        return g3d.ravel()
    else:
        return g3d.reshape(-1, ngrids)

def ifft(g, mesh):
    '''Perform the 3D inverse FFT from reciprocal (G) space to real (R) space.

    Inverse FFT normalization factor is 1./N, same as in `numpy.fft` but
    **different** from MH (they use 1.).

    Args:
        g : (nx*ny*nz,) ndarray
            The function to be inverse FFT'd, flattened to a 1D array
            corresponding to the index order of `span3`.
        mesh : (3,) ndarray of ints (= nx,ny,nz)
            The number G-vectors along each direction.

    Returns:
        (nx*ny*nz,) ndarray
            The inverse FFT 1D array in same index order as Gv (natural order
            of numpy.fft).

    '''
    if g.size == 0:
        return np.zeros_like(g)

    g3d = g.reshape(-1, *mesh)
    assert (g3d.shape[0] == 1 or g[0].size == g3d[0].size)
    f3d = _ifftn_wrapper(g3d)
    ngrids = np.prod(mesh)
    if g.ndim == 1 or (g.ndim == 3 and g.size == ngrids):
        return f3d.ravel()
    else:
        return f3d.reshape(-1, ngrids)


def fftk(f, mesh, expmikr):
    r'''Perform the 3D FFT of a real-space function which is (periodic*e^{ikr}).

    fk(k+G) = \sum_r fk(r) e^{-i(k+G)r} = \sum_r [f(k)e^{-ikr}] e^{-iGr}
    '''
    return fft(f*expmikr, mesh)


def ifftk(g, mesh, expikr):
    r'''Perform the 3D inverse FFT of f(k+G) into a function which is (periodic*e^{ikr}).

    fk(r) = (1/Ng) \sum_G fk(k+G) e^{i(k+G)r} = (1/Ng) \sum_G [fk(k+G)e^{iGr}] e^{ikr}
    '''
    return ifft(g, mesh) * expikr


def get_coulG(cell, k=np.zeros(3), exx=False, mf=None, mesh=None, Gv=None,
              wrap_around=True, omega=None, **kwargs):
    '''Calculate the Coulomb kernel for all G-vectors, handling G=0 and exchange.

    Args:
        k : (3,) ndarray
            k-point
        exx : bool or str
            Whether this is an exchange matrix element
        mf : instance of :class:`SCF`

    Returns:
        coulG : (ngrids,) ndarray
            The Coulomb kernel.
        mesh : (3,) ndarray of ints (= nx,ny,nz)
            The number G-vectors along each direction.
        omega : float
            Enable Coulomb kernel ``erf(|omega|*r12)/r12`` if omega > 0
            and ``erfc(|omega|*r12)/r12`` if omega < 0.
            Note this parameter is slightly different to setting cell.omega for
            exxdiv='ewald' at G0. When cell.omega is configured, the Ewald probe
            charge correction will be computed using the LR or SR Coulomb
            interactions. However, when this kwarg is explicitly specified, the
            exxdiv correction is computed with the full-range Coulomb
            interaction (1/r12). This parameter should only be specified in the
            range-separated JK builder and range-separated DF (and other
            range-separated integral methods if any).
    '''
    exxdiv = exx
    if isinstance(exx, str):
        exxdiv = exx
    elif exx and mf is not None:
        exxdiv = mf.exxdiv

    if mesh is None:
        mesh = cell.mesh
    if 'gs' in kwargs:
        warnings.warn('cell.gs is deprecated.  It is replaced by cell.mesh,'
                      'the number of PWs (=2*gs+1) along each direction.')
        mesh = [2*n+1 for n in kwargs['gs']]
    if Gv is None:
        Gv = cell.get_Gv(mesh)

    if omega is None:
        _omega = cell.omega
    else:
        _omega = omega

    if cell.dimension == 0 and cell.low_dim_ft_type != 'inf_vacuum':
        a = cell.lattice_vectors()
        assert abs(np.eye(3)*a[0,0] - a).max() < 1e-6, \
                'Must be cubic box for cell.dimension=0'
        # ensure the sphere is completely inside the box
        Rc = a[0,0] / 2

        # The truncated Coulomb kernel is
        #    \int_0^R e^{-iG dot r} 1/r dr^3 = 4pi/G \int_0^R sin(G r) dr
        #        = 4pi/G^2 (1 - cos(G R))
        # When R->infinity, the truncated Coulomb becomes the full-range
        # Coulomb. Its kernel is 4pi/G^2. Comparing the two integrals, the
        # divergentintegral
        #    4pi/G^2 \int_R^\inf sin(G R) dr = \int_0^\inf - \int_R^\inf ...
        # would give a regularized value
        #    4pi/G^2 cos (G R)
        # The long-range truncated Coulomb kernel
        #    \int_0^R erf(omega r)/r dr^3 = \int_0^\inf - \int_R^\inf ...
        #        = 4pi/G^2 exp(-G^2/(4 omega^2)) - 4pi/G \int_R^\inf erf(omega r) sin(G r) dr
        # If erf(omega R) ~= 1 for sufficient large R, the second term is
        # simplified to the sin(G r) regularized integral. The long range
        # truncated Coulomb is then given by
        #    4pi/G^2 (exp(-G^2/(4 omega^2) - cos(G R))
        # The short range part is
        #    4pi/G^2 (1 - exp(-G^2/(4 omega^2))
        if (_omega != 0 and
            abs(_omega) * Rc < 2.0): # typically, error of \int erf(omega r) sin (G r) < 1e-5
            raise RuntimeError(
                'In sufficient box size for the truncated range-separated '
                'Coulomb potential in 0D case')

        absG = np.linalg.norm(Gv, axis=1)
        with np.errstate(divide='ignore',invalid='ignore'):
            coulG = 4*np.pi/absG**2
            coulG[0] = 0
        if _omega == 0:
            coulG *= 1. - np.cos(absG*Rc)
            # G=0 term carries the charge. This special term supports the charged
            # system for dimension=0.
            coulG[0] = 2*np.pi*Rc**2
        elif _omega > 0:
            coulG *= np.exp(-.25/_omega**2 * absG**2) - np.cos(absG*Rc)
            coulG[0] = 2*np.pi*Rc**2 - np.pi / _omega**2
        else:
            coulG *= 1 - np.exp(-.25/_omega**2 * absG**2)
            coulG[0] = np.pi / _omega**2
        return coulG

    if abs(k).sum() > 1e-9:
        kG = k + Gv
    else:
        kG = Gv

    equal2boundary = None
    if wrap_around and abs(k).sum() > 1e-9:
        equal2boundary = np.zeros(Gv.shape[0], dtype=bool)
        # Here we 'wrap around' the high frequency k+G vectors into their lower
        # frequency counterparts.  Important if you want the gamma point and k-point
        # answers to agree
        b = cell.reciprocal_vectors()
        box_edge = np.einsum('i,ij->ij', np.asarray(mesh)//2+0.5, b)
        assert (all(np.linalg.solve(box_edge.T, k).astype(int)==0))
        reduced_coords = np.linalg.solve(box_edge.T, kG.T).T
        on_edge_p1 = abs(reduced_coords - 1) < 1e-9
        on_edge_m1 = abs(reduced_coords + 1) < 1e-9
        if cell.dimension >= 1:
            equal2boundary |= on_edge_p1[:,0]
            equal2boundary |= on_edge_m1[:,0]
            kG[reduced_coords[:,0]> 1] -= 2 * box_edge[0]
            kG[reduced_coords[:,0]<-1] += 2 * box_edge[0]
        if cell.dimension >= 2:
            equal2boundary |= on_edge_p1[:,1]
            equal2boundary |= on_edge_m1[:,1]
            kG[reduced_coords[:,1]> 1] -= 2 * box_edge[1]
            kG[reduced_coords[:,1]<-1] += 2 * box_edge[1]
        if cell.dimension == 3:
            equal2boundary |= on_edge_p1[:,2]
            equal2boundary |= on_edge_m1[:,2]
            kG[reduced_coords[:,2]> 1] -= 2 * box_edge[2]
            kG[reduced_coords[:,2]<-1] += 2 * box_edge[2]

    absG2 = np.einsum('gi,gi->g', kG, kG)
    G0_idx = []

    if hasattr(mf, 'kpts'):
        kpts = mf.kpts
    else:
        kpts = k.reshape(1,3)
    Nk = len(kpts)

    if exxdiv == 'vcut_sph':  # PRB 77 193110
        Rc = (3*Nk*cell.vol/(4*np.pi))**(1./3)
        with np.errstate(divide='ignore',invalid='ignore'):
            coulG = 4*np.pi/absG2*(1.0 - np.cos(np.sqrt(absG2)*Rc))
        coulG[absG2==0] = 4*np.pi*0.5*Rc**2

        if cell.dimension < 3:
            raise NotImplementedError

    elif exxdiv == 'vcut_ws':  # PRB 87, 165122
        assert (cell.dimension == 3)
        if not getattr(mf, '_ws_exx', None):
            mf._ws_exx = precompute_exx(cell, kpts)
        exx_alpha = mf._ws_exx['alpha']
        exx_kcell = mf._ws_exx['kcell']
        exx_q = mf._ws_exx['q']
        exx_vq = mf._ws_exx['vq']

        with np.errstate(divide='ignore',invalid='ignore'):
            coulG = 4*np.pi/absG2*(1.0 - np.exp(-absG2/(4*exx_alpha**2)))
        coulG[absG2==0] = np.pi / exx_alpha**2
        # Index k+Gv into the precomputed vq and add on
        gxyz = np.dot(kG, exx_kcell.lattice_vectors().T)/(2*np.pi)
        gxyz = gxyz.round(decimals=6).astype(int)
        mesh = np.asarray(exx_kcell.mesh)
        gxyz = (gxyz + mesh)%mesh
        qidx = (gxyz[:,0]*mesh[1] + gxyz[:,1])*mesh[2] + gxyz[:,2]
        #qidx = [np.linalg.norm(exx_q-kGi,axis=1).argmin() for kGi in kG]
        maxqv = abs(exx_q).max(axis=0)
        is_lt_maxqv = (abs(kG) <= maxqv).all(axis=1)
        coulG = coulG.astype(exx_vq.dtype)
        coulG[is_lt_maxqv] += exx_vq[qidx[is_lt_maxqv]]

        if cell.dimension < 3:
            raise NotImplementedError

    else:
        # Ewald probe charge method to get the leading term of the finite size
        # error in exchange integrals

        G0_idx = np.where(absG2==0)[0]
        if cell.dimension == 3 or cell.low_dim_ft_type == 'inf_vacuum':
            with np.errstate(divide='ignore'):
                coulG = 4*np.pi/absG2
                coulG[G0_idx] = 0

        elif cell.dimension == 2:
            # The following 2D analytical fourier transform is taken from:
            # R. Sundararaman and T. Arias PRB 87, 2013
            b = cell.reciprocal_vectors()
            Ld2 = np.pi/np.linalg.norm(b[2])
            Gz = kG[:,2]
            Gp = np.linalg.norm(kG[:,:2], axis=1)
            weights = 1. - np.cos(Gz*Ld2) * np.exp(-Gp*Ld2)
            with np.errstate(divide='ignore', invalid='ignore'):
                coulG = weights*4*np.pi/absG2
            if len(G0_idx) > 0:
                coulG[G0_idx] = -2*np.pi*Ld2**2 #-pi*L_z^2/2

        elif cell.dimension == 1:
            logger.warn(cell, 'No method for PBC dimension 1, dim-type %s.'
                        '  cell.low_dim_ft_type="inf_vacuum"  should be set.',
                        cell.low_dim_ft_type)
            raise NotImplementedError('truncated coulG for dimension=1 is numerically inaccurate')

            # Carlo A. Rozzi, PRB 73, 205119 (2006)
            a = cell.lattice_vectors()
            # Rc is the cylindrical radius
            Rc = np.sqrt(cell.vol / np.linalg.norm(a[0])) / 2
            Gx = abs(kG[:,0])
            Gp = np.linalg.norm(kG[:,1:], axis=1)
            with np.errstate(divide='ignore', invalid='ignore'):
                weights = 1 + Gp*Rc * scipy.special.j1(Gp*Rc) * scipy.special.k0(Gx*Rc)
                weights -= Gx*Rc * scipy.special.j0(Gp*Rc) * scipy.special.k1(Gx*Rc)
                coulG = 4*np.pi/absG2 * weights
                # TODO: numerical integration
                # coulG[Gx==0] = -4*np.pi * (dr * r * scipy.special.j0(Gp*r) * np.log(r)).sum()
            if len(G0_idx) > 0:
                coulG[G0_idx] = -np.pi*Rc**2 * (2*np.log(Rc) - 1)
        else:
            raise NotImplementedError(f'dimension={cell.dimension} with '
                                      f'low_dim_ft_type={cell.low_dim_ft_type} is not supported')

    if equal2boundary is not None:
        coulG[equal2boundary] = 0

    # Scale the coulG kernel for attenuated Coulomb integrals.
    # * kwarg omega is used by RangeSeparatedJKBuilder which requires ewald probe charge
    # being evaluated with regular Coulomb interaction (1/r12).
    # * cell.omega, which affects the ewald probe charge, is often set by
    # DFT-RSH functionals to build long-range HF-exchange for erf(omega*r12)/r12
    if _omega != 0 and cell.dimension != 3:
        logger.warn(cell, 'The coulG kernel for range-separated Coulomb potential '
                    f'for PBC {cell.dimension} is inaccurate.')
    if _omega > 0:
        # long range part
        coulG *= np.exp(-.25/_omega**2 * absG2)
    elif _omega < 0:
        if exxdiv == 'vcut_sph' or exxdiv == 'vcut_ws':
            raise RuntimeError(f'SR Coulomb for exxdiv={exxdiv} is not available')
        # short range part
        coulG *= (1 - np.exp(-.25/_omega**2 * absG2))

    # For full-range Coulomb and long-range Coulomb,
    # the divergent part of periodic summation of (ii|ii) integrals in
    # Coulomb integrals were cancelled out by electron-nucleus
    # interaction. The periodic part of (ii|ii) in exchange cannot be
    # cancelled out by Coulomb integrals. Its leading term is calculated
    # using Ewald probe charge (the function madelung below)
    if cell.dimension > 0 and exxdiv == 'ewald' and len(G0_idx) > 0:
        if omega is None: # Affects DFT-RSH
            coulG[G0_idx] += Nk*cell.vol*madelung(cell, kpts)
        else: # for RangeSeparatedJKBuilder
            coulG[G0_idx] += Nk*cell.vol*madelung(cell, kpts, omega=0)
    return coulG

def precompute_exx(cell, kpts):
    from pyscf.pbc import gto as pbcgto
    from pyscf.pbc.dft import gen_grid
    log = lib.logger.Logger(cell.stdout, cell.verbose)
    log.debug("# Precomputing Wigner-Seitz EXX kernel")
    Nk = get_monkhorst_pack_size(cell, kpts)
    log.debug("# Nk = %s", Nk)

    kcell = pbcgto.Cell()
    kcell.atom = 'H 0. 0. 0.'
    kcell.spin = 1
    kcell.unit = 'B'
    kcell.verbose = 0
    kcell.a = cell.lattice_vectors() * Nk
    Lc = 1.0/lib.norm(np.linalg.inv(kcell.a), axis=0)
    log.debug("# Lc = %s", Lc)
    Rin = Lc.min() / 2.0
    log.debug("# Rin = %s", Rin)
    # ASE:
    alpha = 5./Rin # sqrt(-ln eps) / Rc, eps ~ 10^{-11}
    log.info("WS alpha = %s", alpha)
    kcell.mesh = np.array([4*int(L*alpha*3.0) for L in Lc])  # ~ [120,120,120]
    # QE:
    #alpha = 3./Rin * np.sqrt(0.5)
    #kcell.mesh = (4*alpha*np.linalg.norm(kcell.a,axis=1)).astype(int)
    log.debug("# kcell.mesh FFT = %s", kcell.mesh)
    rs = kcell.get_uniform_grids(wrap_around=False)
    kngs = len(rs)
    log.debug("# kcell kngs = %d", kngs)
    corners_coord = lib.cartesian_prod(([0, 1], [0, 1], [0, 1]))
    corners = np.dot(corners_coord, kcell.a)
    #vR = np.empty(kngs)
    #for i, rv in enumerate(rs):
    #    # Minimum image convention to corners of kcell parallelepiped
    #    r = lib.norm(rv-corners, axis=1).min()
    #    if np.isclose(r, 0.):
    #        vR[i] = 2*alpha / np.sqrt(np.pi)
    #    else:
    #        vR[i] = scipy.special.erf(alpha*r) / r
    r = np.min([lib.norm(rs-c, axis=1) for c in corners], axis=0)
    vR = scipy.special.erf(alpha*r) / (r+1e-200)
    vR[r<1e-9] = 2*alpha / np.sqrt(np.pi)
    vG = (kcell.vol/kngs) * fft(vR, kcell.mesh)

    if abs(vG.imag).max() > 1e-6:
        # vG should be real in regular lattice. If imaginary part is observed,
        # this probably means a ws cell was built from a unconventional
        # lattice. The SR potential erfc(alpha*r) for the charge in the center
        # of ws cell decays to the region out of ws cell. The Ewald-sum based
        # on the minimum image convention cannot be used to build the kernel
        # Eq (12) of PRB 87, 165122
        raise RuntimeError('Unconventional lattice was found')

    ws_exx = {'alpha': alpha,
              'kcell': kcell,
              'q'    : kcell.Gv,
              'vq'   : vG.real.copy()}
    log.debug("# Finished precomputing")
    return ws_exx


def madelung(cell, kpts, omega=None):
    Nk = get_monkhorst_pack_size(cell, kpts)
    ecell = cell.copy(deep=False)
    ecell._atm = np.array([[1, cell._env.size, 0, 0, 0, 0]])
    ecell._env = np.append(cell._env, [0., 0., 0.])
    ecell.unit = 'B'
    #ecell.verbose = 0
    ecell.a = a = np.einsum('xi,x->xi', cell.lattice_vectors(), Nk)

    if omega is None:
        omega = cell.omega

    if omega == 0:
        return -2*ecell.ewald()

    else:
        # cell.ewald function does not use the Coulomb kernel function
        # get_coulG. When computing the nuclear interactions with attenuated
        # Coulomb operator, the Ewald summation technique is not needed
        # because the Coulomb kernel 4pi/G^2*exp(-G^2/4/omega**2) decays
        # quickly.
        precision = cell.precision
        Ecut = 10.
        Ecut = np.log(16*np.pi**2/(2*omega**2*(2*Ecut)**.5) / precision + 1.) * 2*omega**2
        Ecut = np.log(16*np.pi**2/(2*omega**2*(2*Ecut)**.5) / precision + 1.) * 2*omega**2
        mesh = cutoff_to_mesh(a, Ecut)
        Gv, Gvbase, weights = ecell.get_Gv_weights(mesh)
        wcoulG = get_coulG(ecell, Gv=Gv, omega=abs(omega), exxdiv=None) * weights
        SI = ecell.get_SI(mesh=mesh)
        ZSI = SI[0]
        e_lr = (2*abs(omega)/np.pi**0.5 -
                np.einsum('i,i,i->', ZSI.conj(), ZSI, wcoulG).real)
        if omega > 0:
            return e_lr
        else:
            e_fr = -2*ecell.ewald() # The full-range Coulomb
            return e_fr - e_lr


def get_monkhorst_pack_size(cell, kpts, tol=1e-5):
    kpts = np.reshape(kpts, (-1,3))
    min_tol = tol
    assert kpts.shape[0] < 1/min_tol
    if kpts.shape[0] == 1:
        Nk = np.array([1,1,1])
    else:
        tol = max(10**(-int(-np.log10(1/kpts.shape[0]))-2), min_tol)
        skpts = cell.get_scaled_kpts(kpts)
        Nk = np.array([np.count_nonzero(abs(ski[1:]-ski[:-1]) > tol) + 1
                       for ski in np.sort(skpts.T)])
    return Nk


def get_lattice_Ls(cell, nimgs=None, rcut=None, dimension=None, discard=True):
    '''Get the (Cartesian, unitful) lattice translation vectors for nearby images.
    The translation vectors can be used for the lattice summation.

    Kwargs:
        discard:
            Drop less important Ls based on AO values on grid
    '''
    if dimension is None:
        # For atoms near the boundary of the cell, it is necessary (even in low-
        # dimensional systems) to include lattice translations in all 3 dimensions.
        if cell.dimension < 2 or cell.low_dim_ft_type == 'inf_vacuum':
            dimension = cell.dimension
        else:
            dimension = 3
    if rcut is None:
        rcut = cell.rcut

    if dimension == 0 or rcut <= 0 or cell.natm == 0:
        return np.zeros((1, 3))

    a = cell.lattice_vectors()

    scaled_atom_coords = cell.get_scaled_atom_coords()
    atom_boundary_max = scaled_atom_coords[:,:dimension].max(axis=0)
    atom_boundary_min = scaled_atom_coords[:,:dimension].min(axis=0)
    ovlp_penalty = atom_boundary_max - atom_boundary_min
    dR = ovlp_penalty.dot(a[:dimension])
    dR_basis = np.diag(dR)

    # Search the minimal x,y,z requiring |x*a[0]+y*a[1]+z*a[2]+dR|^2 > rcut^2
    # Ls boundary should be derived by decomposing (a, Rij) for each atom-pair.
    # For reasons unclear, the so-obtained Ls boundary seems not large enough.
    # The upper-bound of the Ls boundary is generated by find_boundary function.
    def find_boundary(a):
        aR = np.vstack([a, dR_basis])
        r = np.linalg.qr(aR.T)[1]
        ub = (rcut + abs(r[2,3:]).sum()) / abs(r[2,2])
        return ub

    xb = find_boundary(a[[1,2,0]])
    if dimension > 1:
        yb = find_boundary(a[[2,0,1]])
    else:
        yb = 0
    if dimension > 2:
        zb = find_boundary(a)
    else:
        zb = 0
    bounds = np.ceil([xb, yb, zb]).astype(int)
    Ts = lib.cartesian_prod((np.arange(-bounds[0], bounds[0]+1),
                             np.arange(-bounds[1], bounds[1]+1),
                             np.arange(-bounds[2], bounds[2]+1)))
    Ls = np.dot(Ts[:,:dimension], a[:dimension])

    if discard and len(Ls) > 1:
        r = cell.atom_coords()
        rr = r[:,None] - r
        dist_max = np.linalg.norm(rr, axis=2).max()
        Ls_mask = np.linalg.norm(Ls, axis=1) < rcut + dist_max
        Ls = Ls[Ls_mask]
    return np.asarray(Ls, order='C')

def check_lattice_sum_range(cell, Ls):
    '''
    Evaluates whether the lattice summation range is sufficient.

    This function calculates the minimum distance between atoms in the primary
    unit cell and atoms in lattice images *not* included in the specified
    lattice sum vectors (Ls).
    '''
    Ls_full = get_lattice_Ls(cell, rcut=cell.rcut*1.5, discard=False)
    Ls_idx = intersection(Ls_full, Ls)
    Ls_remaining = np.setdiff1d(np.arange(len(Ls_full)), Ls_idx)
    atom_coords = cell.atom_coords()
    atoms_outside = (Ls_full[Ls_remaining,None] + atom_coords).reshape(-1, 3)
    return np.linalg.norm(atoms_outside[:,None] - atom_coords, axis=2).min()

def super_cell(cell, ncopy, wrap_around=False):
    '''Create an ncopy[0] x ncopy[1] x ncopy[2] supercell of the input cell
    Note this function differs from :func:`cell_plus_imgs` that cell_plus_imgs
    creates images in both +/- direction.

    Args:
        cell : instance of :class:`Cell`

        ncopy : (3,) array

        wrap_around : bool
            Put the original cell centered on the super cell. It has the
            effects corresponding to the parameter wrap_around of
            cell.make_kpts.

    Returns:
        supcell : instance of :class:`Cell`
    '''
    a = cell.lattice_vectors()
    #:supcell.atom = []
    #:for Lx in range(ncopy[0]):
    #:    for Ly in range(ncopy[1]):
    #:        for Lz in range(ncopy[2]):
    #:            # Using cell._atom guarantees coord is in Bohr
    #:            for atom, coord in cell._atom:
    #:                L = np.dot([Lx, Ly, Lz], a)
    #:                supcell.atom.append([atom, coord + L])
    xs = np.arange(ncopy[0])
    ys = np.arange(ncopy[1])
    zs = np.arange(ncopy[2])
    if wrap_around:
        xs[(ncopy[0]+1)//2:] -= ncopy[0]
        ys[(ncopy[1]+1)//2:] -= ncopy[1]
        zs[(ncopy[2]+1)//2:] -= ncopy[2]
    Ts = lib.cartesian_prod((xs, ys, zs))
    Ls = np.dot(Ts, a)
    supcell = cell.copy(deep=False)
    supcell.a = np.einsum('i,ij->ij', ncopy, a)
    mesh = np.asarray(ncopy) * np.asarray(cell.mesh)
    supcell.mesh = (mesh // 2) * 2 + 1
    if isinstance(cell.magmom, np.ndarray):
        supcell.magmom = cell.magmom.tolist() * np.prod(ncopy)
    else:
        supcell.magmom = cell.magmom * np.prod(ncopy)
    return _build_supcell_(supcell, cell, Ls)


def cell_plus_imgs(cell, nimgs):
    '''Create a supercell via nimgs[i] in each +/- direction, as in get_lattice_Ls().
    Note this function differs from :func:`super_cell` that super_cell only
    stacks the images in + direction.

    Args:
        cell : instance of :class:`Cell`
        nimgs : (3,) array

    Returns:
        supcell : instance of :class:`Cell`
    '''
    a = cell.lattice_vectors()
    Ts = lib.cartesian_prod((np.arange(-nimgs[0], nimgs[0]+1),
                             np.arange(-nimgs[1], nimgs[1]+1),
                             np.arange(-nimgs[2], nimgs[2]+1)))
    Ls = np.dot(Ts, a)
    supcell = cell.copy(deep=False)
    supcell.a = np.einsum('i,ij->ij', nimgs, a)
    supcell.mesh = np.array([(nimgs[0]*2+1)*cell.mesh[0],
                             (nimgs[1]*2+1)*cell.mesh[1],
                             (nimgs[2]*2+1)*cell.mesh[2]])
    return _build_supcell_(supcell, cell, Ls)

def _build_supcell_(supcell, cell, Ls):
    '''
    Construct supcell ._env directly without calling supcell.build() method.
    This reserves the basis contraction coefficients defined in cell
    '''
    from pyscf.pbc import gto as pbcgto
    nimgs = len(Ls)
    symbs = [atom[0] for atom in cell._atom] * nimgs
    coords = Ls.reshape(-1,1,3) + cell.atom_coords()
    coords = coords.reshape(-1,3)
    x, y, z = coords.T
    supcell.atom = supcell._atom = list(zip(symbs, zip(x, y, z)))
    supcell.unit = 'B'
    supcell.enuc = None # reset nuclear energy

    # Do not call supcell.build() to initialize supcell since it may normalize
    # the basis contraction coefficients

    # preserves environments defined in cell._env (e.g. omega, gauge origin)
    _env = np.append(cell._env, coords.ravel())
    _atm = np.repeat(cell._atm[None,:,:], nimgs, axis=0)
    _atm = _atm.reshape(-1, ATM_SLOTS)
    # Point to the coordinates appended to _env
    _atm[:,PTR_COORD] = cell._env.size + np.arange(nimgs * cell.natm) * 3

    _bas = np.repeat(cell._bas[None,:,:], nimgs, axis=0)
    # For atom pointers in each image, shift natm*image_id
    _bas[:,:,ATOM_OF] += np.arange(nimgs)[:,None] * cell.natm

    supcell._atm = np.asarray(_atm, dtype=np.int32)
    supcell._bas = np.asarray(_bas.reshape(-1, BAS_SLOTS), dtype=np.int32)
    supcell._env = _env

    if isinstance(supcell, pbcgto.Cell) and getattr(supcell, 'space_group_symmetry', False):
        supcell.build_lattice_symmetry(not cell._mesh_from_build)
    return supcell


def cutoff_to_mesh(a, cutoff):
    r'''
    Convert KE cutoff to FFT-mesh

        uses KE = k^2 / 2, where k_max ~ \pi / grid_spacing

    Args:
        a : (3,3) ndarray
            The real-space cell lattice vectors. Each row represents a
            lattice vector.
        cutoff : float
            KE energy cutoff in a.u.

    Returns:
        mesh : (3,) array
    '''
    # Search the minimal x,y,z requiring |x*b[0]+y*b[1]+z*b[2]|^2 > 2 * cutoff
    b = 2 * np.pi * np.linalg.inv(a.T)
    rx = np.linalg.qr(b[[1,2,0]].T)[1][2,2]
    ry = np.linalg.qr(b[[2,0,1]].T)[1][2,2]
    rz = np.linalg.qr(b.T)[1][2,2]

    Gmax = (2*cutoff)**.5 / np.abs([rx, ry, rz])
    mesh = np.ceil(Gmax).astype(int) * 2 + 1
    return mesh

def mesh_to_cutoff(a, mesh):
    '''
    Convert #grid points to KE cutoff
    '''
    # Search the minimal x,y,z requiring |x*b[0]+y*b[1]+z*b[2]|^2 > 2 * cutoff
    b = 2 * np.pi * np.linalg.inv(a.T)
    rx = np.linalg.qr(b[[1,2,0]].T)[1][2,2]
    ry = np.linalg.qr(b[[2,0,1]].T)[1][2,2]
    rz = np.linalg.qr(b.T)[1][2,2]

    gs = (np.asarray(mesh) - 1) // 2
    Gmax = gs * np.array([rx, ry, rz])
    ke_cutoff = Gmax**2 / 2
    return ke_cutoff

def cutoff_to_gs(a, cutoff):
    '''Deprecated.  Replaced by function cutoff_to_mesh.'''
    return [n//2 for n in cutoff_to_mesh(a, cutoff)]

def gs_to_cutoff(a, gs):
    '''Deprecated.  Replaced by function mesh_to_cutoff.'''
    return mesh_to_cutoff(a, [2*n+1 for n in gs])

def round_to_cell0(r, tol=1e-6):
    '''Round scaled coordinates to reference unit cell
    '''
    from pyscf.pbc.lib import kpts_helper
    return kpts_helper.round_to_fbz(r, wrap_around=False, tol=tol)



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pyscf.mcscf import newton_casscf, casci, mc1step
from pyscf.grad import rks as rks_grad
from pyscf.dft import gen_grid
from pyscf.lib import logger, pack_tril, current_memory, einsum, tag_array
from pyscf.grad import sacasscf
from pyscf.mcscf.casci import cas_natorb

from pyscf.mcpdft.pdft_eff import _contract_eff_rho
from pyscf.mcpdft.otpd import get_ontop_pair_density, _grid_ao2mo
from pyscf.mcpdft import _dms

from itertools import product
from scipy import linalg
import numpy as np
import gc

BLKSIZE = gen_grid.BLKSIZE

def gfock_sym(mc, mo_coeff, casdm1, casdm2, h1e, eris):
    """Assume that h2e v_j = v_k"""
    ncore = mc.ncore
    ncas = mc.ncas
    nocc = ncore + ncas

    nao, nmo = mo_coeff.shape

    # gfock = Generalized Fock, Adv. Chem. Phys., 69, 63

    # MRH: I need to replace aapa with the equivalent array from veff2
    # I'm not sure how the outcore file-paging system works
    # I also need to generate vhf_c and vhf_a from veff2 rather than the
    # molecule's actual integrals. The true Coulomb repulsion should already be
    # in veff1, but I need to generate the "fake" vj - vk/2 from veff2
    h1e_mo = mo_coeff.T @ h1e @ mo_coeff + eris.vhf_c
    aapa = np.zeros((ncas, ncas, nmo, ncas), dtype=h1e_mo.dtype)
    vhf_a = np.zeros((nmo, nmo), dtype=h1e_mo.dtype)

    for i in range(nmo):
        jbuf = eris.ppaa[i]
        aapa[:, :, i, :] = jbuf[ncore:nocc, :, :]
        vhf_a[i] = np.tensordot(jbuf, casdm1, axes=2)

    vhf_a *= 0.5
    # we have assumed that vj = vk: vj - vk/2 = vj - vj/2 = vj/2
    gfock = np.zeros((nmo, nmo))
    gfock[:, :ncore] = (h1e_mo[:, :ncore] + vhf_a[:, :ncore]) * 2
    gfock[:, ncore:nocc] = h1e_mo[:, ncore:nocc] @ casdm1
    gfock[:, ncore:nocc] += einsum('uviw,vuwt->it', aapa, casdm2)

    return gfock

def xc_response(ot, vot, rho, Pi, weights, moval_occ, aoval, mo_occ, mo_occup, ncore, nocc, casdm2_pack, ndpi, mo_cas):
    vrho, vPi = vot


    # Vpq + Vpqrs * Drs ; I'm not sure why the list comprehension down
    # there doesn't break ao's stride order but I'm not complaining
    vrho = _contract_eff_rho(vPi, rho.sum(0), add_eff_rho=vrho)

    tmp_dv = np.stack(
        [
            ot.get_veff_1body(rho, Pi, [ao_i, moval_occ], weights, kern=vrho)
            for ao_i in aoval
        ],
        axis=0,
    )
    tmp_dv = (tmp_dv * mo_occ[None, :, :] * mo_occup[None, None, :nocc]).sum(2)

    # Vpuvx * Lpuvx ; remember the stupid slowest->fastest->medium
    # stride order of the ao grid arrays
    moval_cas = np.ascontiguousarray(moval_occ[..., ncore:].transpose(0,2,1)).transpose(0,2,1)

    tmp_dv1 = ot.get_veff_2body_kl(rho, Pi, moval_cas, moval_cas, weights, symm=True, kern=vPi)
    # tmp_dv.shape = ndpi,ngrids,ncas*(ncas+1)//2
    tmp_dv1 = np.tensordot(tmp_dv1, casdm2_pack, axes=(-1,-1))
    # tmp_dv.shape = ndpi, ngrids, ncas, ncas
    tmp_dv1[0] = (tmp_dv1[:ndpi] * moval_cas[:ndpi, :, None, :]).sum(0)
    # Chain and product rule
    tmp_dv1[1:ndpi] *= moval_cas[0, :, None, :]
    # Chain and product rule
    tmp_dv1 = tmp_dv1.sum(-1)
    # tmp_dv.shape = ndpi, ngrids, ncas
    tmp_dv1 = np.tensordot(aoval[:, :ndpi], tmp_dv1, axes=((1, 2), (0, 1)))
    # tmp_dv.shape = comp, nao (orb), ncas (dm2)
    tmp_dv1 = np.einsum('cpu,pu->cp', tmp_dv1, mo_cas)
    # tmp_dv.shape = comp, ncas

    return tmp_dv + tmp_dv1


def pack_casdm2(cascm2, ncas):
    diag_idx = np.arange(ncas)  # for puvx
    diag_idx = diag_idx * (diag_idx+1) // 2 + diag_idx

    casdm2_pack = (cascm2 + cascm2.transpose(0, 1, 3, 2)).reshape(ncas**2, ncas, ncas)
    casdm2_pack = pack_tril(casdm2_pack).reshape(ncas, ncas, -1)
    casdm2_pack[:, :, diag_idx] *= 0.5
    return casdm2_pack

def sum_terms(mf_grad, mol, atmlst,dm1, gfock, coul_term, dvxc):
    de_hcore = np.zeros((len(atmlst), 3))
    de_renorm = np.zeros((len(atmlst), 3))
    de_coul = np.zeros((len(atmlst), 3))
    de_xc = np.zeros((len(atmlst), 3))

    aoslices = mol.aoslice_by_atom()
    hcore_deriv = mf_grad.hcore_generator(mol)
    s1 = mf_grad.get_ovlp(mol)

    for k, ia in enumerate(atmlst):
        p0, p1 = aoslices[ia][2:]
        h1ao = hcore_deriv(ia)
        de_hcore[k] += np.tensordot(h1ao, dm1)
        de_renorm[k] -= np.tensordot(s1[:, p0:p1], gfock[p0:p1])*2
        de_coul[k] += coul_term(p0, p1)
        de_xc[k] += dvxc[:, p0:p1].sum(1)*2

    de_nuc = mf_grad.grad_nuc(mol, atmlst)

    return de_hcore, de_coul, de_xc, de_nuc, de_renorm,

def mcpdft_HellmanFeynman_grad (mc, ot, veff1, veff2, mo_coeff=None, ci=None,
        atmlst=None, mf_grad=None, verbose=None, max_memory=None,
        auxbasis_response=False):
    '''Modification of pyscf.grad.casscf.kernel to compute instead the
    Hellman-Feynman gradient terms of MC-PDFT. From the differentiated
    Hamiltonian matrix elements, only the core and Coulomb energy parts
    remain. For the renormalization terms, the effective Fock matrix is
    as in CASSCF, but with the same Hamiltonian substutition that is
    used for the energy response terms. '''
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if mf_grad is None: mf_grad = mc.get_rhf_base ().nuc_grad_method()
    if mc.frozen is not None:
        raise NotImplementedError
    if max_memory is None: max_memory = mc.max_memory
    t0 = (logger.process_clock (), logger.perf_counter ())

    mol = mc.mol
    ncore = mc.ncore
    ncas = mc.ncas
    nocc = ncore + ncas
    nelecas = mc.nelecas
    nao, nmo = mo_coeff.shape

    mo_core = mo_coeff[:,:ncore]
    mo_cas = mo_coeff[:,ncore:nocc]

    casdm1, casdm2 = mc.fcisolver.make_rdm12(ci, ncas, nelecas)

    spin = abs(nelecas[0] - nelecas[1])
    omega, _, hyb = ot._numint.rsh_and_hybrid_coeff(ot.otxc, spin=spin)
    if abs(omega) > 1e-11:
        raise NotImplementedError("range-separated on-top functionals")
    if abs(hyb[0] - hyb[1]) > 1e-11:
        raise NotImplementedError(
            "hybrid on-top functionals with different exchange,correlation components"
        )

    cas_hyb = hyb[0]
    ot_hyb = 1.0 - cas_hyb

    if cas_hyb > 1e-11:
        # TODO: actually implement this in a more efficient manner
        # That is, lets not call grad_elec, but lets actually do this our self maybe?
        # Can then use get_pdft_veff with drop_mcwfn = False to automatically get
        # Generalized fock matrix terms, but then have to deal explicitly with the
        # derivative of eri terms and auxbasis stuff. Also, get_pdft_veff with
        # drop_mcwfn=False means the get_wfn_response doesn't have to add the
        # eris to the veff2 since it should just include it already
        if auxbasis_response:
            from pyscf.df.grad import casscf as casscf_grad
        else:
            from pyscf.grad import casscf as casscf_grad

        cas_grad = casscf_grad.Gradients(mc)

        de_cas = cas_hyb * cas_grad.grad_elec(
            mo_coeff=mo_coeff, ci=ci, atmlst=atmlst, verbose=verbose
        )

    # gfock = Generalized Fock, Adv. Chem. Phys., 69, 63
    dm_core = 2 * mo_core @ mo_core.T
    dm_cas = mo_cas @ casdm1 @ mo_cas.T

    gfock = gfock_sym(mc, mo_coeff, casdm1, casdm2, ot_hyb*mc.get_hcore() + veff1, veff2)

    dme0 = mo_coeff @ (0.5*(gfock+gfock.T)) @ mo_coeff.T
    del gfock

    if atmlst is None:
        atmlst = range(mol.natm)

    de_grid = np.zeros ((len(atmlst),3))
    de_wgt = np.zeros ((len(atmlst),3))
    de_aux = np.zeros ((len(atmlst),3))

    t0 = logger.timer (mc, 'PDFT HlFn gfock', *t0)
    mo_coeff, ci, mo_occup = cas_natorb (mc, mo_coeff=mo_coeff, ci=ci)
    mo_occ = mo_coeff[:,:nocc]
    mo_cas = mo_coeff[:,ncore:nocc]

    dm1 = dm_core + dm_cas
    dm1 = tag_array (dm1, mo_coeff=mo_coeff, mo_occ=mo_occup)

    # MRH: vhf1c and vhf1a should be the TRUE vj_c and vj_a (no vk!)
    vj = mf_grad.get_jk (dm=dm1)[0]
    if auxbasis_response:
        de_aux += ot_hyb*np.squeeze (vj.aux[:,:,atmlst,:])

    # MRH: Now I have to compute the gradient of the on-top energy
    # This involves derivatives of the orbitals that construct rho and Pi and
    # therefore another set of potentials. It also involves the derivatives of
    # quadrature grid points which propagate through the densities and
    # therefore yet another set of potentials. The orbital-derivative part
    # includes all the grid points and some of the orbitals (- sign); the
    # grid-derivative part includes all of the orbitals and some of the grid
    # points (+ sign). I'll do a loop over grid sections and make arrays of
    # type (3,nao, nao) and (3,nao, ncas, ncas, ncas). I'll contract them
    # within the grid loop for the grid derivatives and in the following
    # orbital loop for the xc derivatives
    # MRH, 05/09/2020: The actual spin density doesn't matter at all in PDFT!
    # I could probably save a fair amount of time by not screwing around with
    # the actual spin density! Also, the cumulant decomposition can always be
    # defined without the spin-density matrices and it's still valid!
    casdm1, casdm2 = mc.fcisolver.make_rdm12(ci, ncas, nelecas)
    twoCDM = _dms.dm2_cumulant (casdm2, casdm1)
    dm1 = tag_array (dm1, mo_coeff=mo_occ, mo_occ=mo_occup[:nocc])
    make_rho = ot._numint._gen_rho_evaluator (mol, dm1, hermi=1, with_lapl=False)[0]
    dvxc = np.zeros ((3,nao))
    idx = np.array ([[1,4,5,6],[2,5,7,8],[3,6,8,9]], dtype=np.int_)
    # For addressing particular ao derivatives
    if ot.xctype == 'LDA': idx = idx[:,0:1] # For LDAs, no second derivatives

    casdm2_pack = pack_casdm2(twoCDM, ncas)
    full_atmlst = -np.ones (mol.natm, dtype=np.int_)

    t1 = logger.timer (mc, 'PDFT HlFn quadrature setup', *t0)
    for k, ia in enumerate (atmlst):
        full_atmlst[ia] = k

    # for LDA we need 1 deriv, GGA: 2 deriv
    # for mGGA with tau, we only need 2 deriv
    ao_deriv = (1, 2, 2)[ot.dens_deriv]
    ndao = (1, 4, 10)[ao_deriv]
    ndrho = (1, 4, 5)[ot.dens_deriv]
    ndpi = (1, 4)[ot.Pi_deriv]
    ncols = 1.05 * 3 * (ndao * (nao + nocc) + max(ndrho * nao, ndpi * ncas * ncas))

    for ia, (coords, w0, w1) in enumerate (rks_grad.grids_response_cc (
            ot.grids)):
        # For the xc potential derivative, I need every grid point in the
        # entire molecule regardless of atmlist. (Because that's about orbs.)
        # For the grid and weight derivatives, I only need the gridpoints that
        # are in atmlst. It is conceivable that I can make this more efficient
        # by only doing cross-combinations of grids and AOs, but I don't know
        # how "mask" works yet or how else I could do this.
        gc.collect ()
        ngrids = coords.shape[0]

        remaining_floats = (max_memory - current_memory()[0]) * 1e6 / 8
        blksize = int (remaining_floats / (ncols*BLKSIZE)) * BLKSIZE
        blksize = max (BLKSIZE, min (blksize, ngrids, BLKSIZE*1200))
        t1 = logger.timer (mc, 'PDFT HlFn quadrature atom {} mask and memory '
            'setup'.format (ia), *t1)

        for ip0 in range(0, ngrids, blksize):
            ip1 = min(ngrids, ip0 + blksize)
            mask = gen_grid.make_mask(mol, coords[ip0:ip1])
            logger.info(
                mc,
                ("PDFT gradient atom {} slice {}-{} of {} total").format(
                    ia, ip0, ip1, ngrids
                ),
            )

            ao = ot._numint.eval_ao(mol, coords[ip0:ip1], deriv=ao_deriv, non0tab=mask)

            # Need 1st derivs for LDA, 2nd for GGA, etc.
            t1 = logger.timer(
                mc, ("PDFT HlFn quadrature atom {} ao " "grids").format(ia), *t1
            )
            # Slice down ao so as not to confuse the rho and Pi generators
            if ot.xctype == "LDA":
                aoval = ao[0]
            elif ot.xctype == "GGA":
                aoval = ao[:4]
            elif ot.xctype == "MGGA":
                aoval = ao[:4]
            else:
                raise ValueError("Unknown xctype: {}".format(ot.xctype))

            rho = make_rho (0, aoval, mask, ot.xctype) / 2.0
            rho = np.stack ((rho,)*2, axis=0)
            t1 = logger.timer (mc, ('PDFT HlFn quadrature atom {} rho '
                'calc').format (ia), *t1)
            Pi = get_ontop_pair_density (ot, rho, aoval, twoCDM, mo_cas,
                ot.Pi_deriv, mask)
            t1 = logger.timer (mc, ('PDFT HlFn quadrature atom {} Pi '
                'calc').format (ia), *t1)

            # TODO: consistent format requirements for shape of ao grid
            if ot.xctype == 'LDA':
                aoval = ao[:1]
            moval_occ = _grid_ao2mo (mol, aoval, mo_occ, mask)
            t1 = logger.timer (mc, ('PDFT HlFn quadrature atom {} ao2mo '
                'grid').format (ia), *t1)
            aoval = np.ascontiguousarray ([ao[ix].transpose (0,2,1)
                for ix in idx[:,:ndao]]).transpose (0,1,3,2)
            ao = None
            t1 = logger.timer (mc, ('PDFT HlFn quadrature atom {} ao grid '
                'reshape').format (ia), *t1)
            eot, vot = ot.eval_ot (rho, Pi, weights=w0[ip0:ip1])[:2]
            t1 = logger.timer (mc, ('PDFT HlFn quadrature atom {} '
                'eval_ot').format (ia), *t1)
            puvx_mem = 2 * ndpi * (ip1-ip0) * ncas * ncas * 8 / 1e6
            remaining_mem = max_memory - current_memory ()[0]
            logger.info (mc, ('PDFT gradient memory note: working on {} grid '
                'points; estimated puvx usage = {:.1f} of {:.1f} remaining '
                'MB').format ((ip1-ip0), puvx_mem, remaining_mem))

            # Weight response
            de_wgt += np.tensordot (eot, w1[atmlst,...,ip0:ip1], axes=(0,2))
            t1 = logger.timer (mc, ('PDFT HlFn quadrature atom {} weight '
                'response').format (ia), *t1)

            # Find the atoms that are a part of the atomlist
            # grid correction shouldn't be added if they aren't there
            # The last stuff to vectorize is in get_veff_2body!
            k = full_atmlst[ia]

            tmp_dv = xc_response(ot, vot, rho, Pi, w0[ip0:ip1], moval_occ, aoval, mo_occ, mo_occup, ncore, nocc,
                                 casdm2_pack, ndpi, mo_cas)

            if k >=0: de_grid[k] += 2*tmp_dv.sum(1) # Grid response
            dvxc -= tmp_dv #XC response

            tmp_dv = None
            t1 = logger.timer (mc, ('PDFT HlFn quadrature atom {}').format (ia), *t1)

            rho = Pi = eot = vot = aoval = moval_occ = None
            gc.collect ()

    def coul_term(p0, p1):
        return np.tensordot(vj[:,p0:p1], dm1[p0:p1])*2

    de_hcore, de_coul, de_xc, de_nuc, de_renorm = sum_terms(mf_grad, mol, atmlst, dm1, dme0, coul_term,
                                                                        dvxc)
    # Deal with hybridization
    de_hcore *= ot_hyb
    de_coul *= ot_hyb

    logger.debug (mc, "MC-PDFT Hellmann-Feynman nuclear:\n{}".format (de_nuc))
    logger.debug (mc, "MC-PDFT Hellmann-Feynman hcore component:\n{}".format (
        de_hcore))
    logger.debug (mc, "MC-PDFT Hellmann-Feynman coulomb component:\n{}".format
        (de_coul))
    logger.debug (mc, "MC-PDFT Hellmann-Feynman xc component:\n{}".format (
        de_xc))
    logger.debug (mc, ("MC-PDFT Hellmann-Feynman quadrature point component:"
        "\n{}").format (de_grid))
    logger.debug (mc, ("MC-PDFT Hellmann-Feynman quadrature weight component:"
        "\n{}").format (de_wgt))
    logger.debug (mc, "MC-PDFT Hellmann-Feynman renorm component:\n{}".format (
        de_renorm))

    de = de_nuc + de_hcore + de_coul + de_renorm + de_xc + de_grid + de_wgt


    if auxbasis_response:
        de += de_aux
        logger.debug (mc, "MC-PDFT Hellmann-Feynman aux component:\n{}".format
            (de_aux))

    if cas_hyb > 1e-11:
        de += de_cas
        logger.debug(mc, "MC-PDFT Hellmann-Feynman CAS component:\n{}".format(de_cas))

    t1 = logger.timer (mc, 'PDFT HlFn total', *t0)

    return de

# TODO: docstrings (parent classes???)
# TODO: add a consistent threshold for elimination of degenerate-state rotations
class Gradients (sacasscf.Gradients):

    def __init__(self, pdft, state=None):
        super().__init__(pdft, state=state)
        # TODO: gradient of PDFT state-average energy
        # i.e., state = 0 & nroots > 1 case
        if self.state is None and self.nroots == 1:
            self.state = 0
        self.e_mcscf = self.base.e_mcscf
        self._not_implemented_check ()

    def _not_implemented_check (self):
        name = self.__class__.__name__
        if (isinstance (self.base, casci.CASCI) and not
            isinstance (self.base, mc1step.CASSCF)):
            raise NotImplementedError (
                "{} for CASCI-based MC-PDFT".format (name)
            )
        ot, otxc, nelecas = self.base.otfnal, self.base.otxc, self.base.nelecas
        spin = abs (nelecas[0]-nelecas[1])
        omega, alpha, hyb = ot._numint.rsh_and_hybrid_coeff (
            otxc, spin=spin)
        hyb_x, hyb_c = hyb
        if abs(hyb_x - hyb_c) >1e-11:
            raise NotImplementedError (
                "{} for hybrid MC-PDFT functionals with different exchange, correlation".format (name)
            )
        if omega:
            raise NotImplementedError (
                "{} for range-separated MC-PDFT functionals".format (name)
            )

    def get_wfn_response (self, state=None, verbose=None, mo=None,
            ci=None, veff1=None, veff2=None, nlag=None, **kwargs):
        if state is None: state = self.state
        if verbose is None: verbose = self.verbose
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if nlag is None: nlag = self.nlag
        if (veff1 is None) or (veff2 is None):
            veff1, veff2 = self.base.get_pdft_veff (mo, ci[state],
                incl_coul=True, paaa_only=True, drop_mcwfn=True)

        log = logger.new_logger(self, verbose)

        sing_tol = getattr(self, 'sing_tol_sasa', 1e-8)
        fcasscf = self.make_fcasscf(state)
        fcasscf.mo_coeff = mo
        fcasscf.ci = ci[state]
        def my_hcore ():
            return self.base.get_hcore () + veff1
        fcasscf.get_hcore = my_hcore

        nelecas = self.base.nelecas
        spin = abs (nelecas[0]-nelecas[1])
        omega, alpha, hyb = self.base.otfnal._numint.rsh_and_hybrid_coeff(self.base.otxc, spin=spin)
        if omega:
            raise NotImplementedError("range-separated MC-PDFT functionals")
        if abs(hyb[0] - hyb[1]) > 1e-11:
            raise NotImplementedError("hybrid on-top functional with different exchange,correlation components")
        cas_hyb = hyb[0]

        if cas_hyb > 1e-11:# and len(ci) > 1:
            # For SS-CASSCF, there are no Lagrange multipliers
            # This is only needed in SA case
            eris = self.base.ao2mo(mo_coeff=mo)
            terms = ["vhf_c", "papa", "ppaa", "j_pc", "k_pc"]
            for term in terms:
                setattr(eris, term, getattr(veff2, term) + cas_hyb*getattr(eris, term)[:])
            veff2.vhf_c
            veff2 = eris


        g_all_state = newton_casscf.gen_g_hop (fcasscf, mo, ci[state], veff2, verbose)[0]

        g_all = np.zeros (nlag)
        g_all[:self.ngorb] = g_all_state[:self.ngorb]
        # Eliminate gradient of self-rotation and rotation into
        # degenerate states
        spin_states = np.asarray (self.spin_states)
        idx_spin = spin_states==spin_states[state]
        e_gap = self.e_mcscf-self.e_mcscf[state] if self.nroots>1 else [0.0]
        idx_degen = np.abs (e_gap)<sing_tol
        idx = np.where (idx_spin & idx_degen)[0]
        assert (state in idx)
        gci_state = g_all_state[self.ngorb:]
        ci_proj = np.asarray ([ci[i].ravel () for i in idx])
        gci_sa = np.dot (ci_proj, gci_state)
        gci_state -= np.dot (gci_sa, ci_proj)
        gci = g_all[self.ngorb:]
        offs = 0
        if state>0:
            offs = sum ([na * nb for na, nb in zip(
                        self.na_states[:state], self.nb_states[:state])])
        ndet = self.na_states[state]*self.nb_states[state]
        gci[offs:][:ndet] += gci_state

        # Debug
        log.debug("g_all mo:\n{}".format(g_all[:self.ngorb]))
        log.debug("g_all CI:\n{}".format(g_all[self.ngorb:]))

        return g_all

    def get_ham_response (self, state=None, atmlst=None, verbose=None, mo=None,
            ci=None, eris=None, mf_grad=None, veff1=None, veff2=None,
            **kwargs):
        if state is None: state = self.state
        if atmlst is None: atmlst = self.atmlst
        if verbose is None: verbose = self.verbose
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if (veff1 is None) or (veff2 is None):
            assert (False), kwargs
        fcasscf = self.make_fcasscf (state)
        fcasscf.mo_coeff = mo
        fcasscf.ci = ci[state]

        return mcpdft_HellmanFeynman_grad(
            fcasscf,
            self.base.otfnal,
            veff1,
            veff2,
            mo_coeff=mo,
            ci=ci[state],
            atmlst=atmlst,
            mf_grad=mf_grad,
            verbose=verbose,
        )

    def get_init_guess (self, bvec, Adiag, Aop, precond):
        '''Initial guess should solve the problem for SA-SA rotations'''
        sing_tol = getattr (self, 'sing_tol_sasa', 1e-8)
        ci = self.base.ci
        state = self.state
        if self.nroots == 1: ci = [ci,]
        idx_spin = [i for i in range (self.nroots)
                    if self.spin_states[i]==self.spin_states[state]]
        ci_blk = np.asarray ([ci[i].ravel () for i in idx_spin])
        b_orb, b_ci = self.unpack_uniq_var (bvec)
        b_ci_blk = np.asarray ([b_ci[i].ravel () for i in idx_spin])
        x0 = np.zeros_like (bvec)
        if self.nroots > 1:
            b_sa = np.dot (ci_blk.conjugate (), b_ci[state].ravel ())
            A_sa = 2 * self.weights[state] * (self.e_mcscf
                - self.e_mcscf[state])
            idx_null = np.abs (A_sa)<sing_tol
            assert (idx_null[state])
            A_sa = A_sa[idx_spin]
            idx_null = np.abs (A_sa)<sing_tol
            if np.any (np.abs (b_sa[idx_null])>=sing_tol):
                logger.warn (self, 'Singular Hessian in CP-MCPDFT!')
            idx_null &= np.abs (b_sa)<sing_tol
            A_sa[idx_null] = sing_tol
            x0_sa = -b_sa / A_sa # Hessian is diagonal so: easy
            ovlp = ci_blk.conjugate () @ b_ci_blk.T
            logger.debug (self, 'Linear response SA-SA part:\n{}'.format (
                ovlp))
            logger.debug (self, 'Linear response SA-CI norms:\n{}'.format (
                linalg.norm (b_ci_blk.T - ci_blk.T @ ovlp, axis=1)))
            if self.ngorb: logger.debug (self, 'Linear response orbital '
                'norms:\n{}'.format (linalg.norm (bvec[:self.ngorb])))
            logger.debug (self, 'SA-SA Lagrange multiplier for root '
                '{}:\n{}'.format (state, x0_sa))
            x0_orb, x0_ci = self.unpack_uniq_var (x0)
            x0_ci[state] = np.dot (x0_sa, ci_blk).reshape (
                self.na_states[state], self.nb_states[state])
            x0 = self.pack_uniq_var (x0_orb, x0_ci)
        r0 = bvec + Aop (x0)
        r0_orb, r0_ci = self.unpack_uniq_var (r0)
        r0_ci_blk = np.asarray ([r0_ci[i].ravel () for i in idx_spin])
        ovlp = ci_blk.conjugate () @ r0_ci_blk.T
        logger.debug (self, 'Lagrange residual SA-SA part after solving SA-SA'
            ' part:\n{}'.format (ovlp))
        logger.debug (self, 'Lagrange residual SA-CI norms after solving SA-SA'
            ' part:\n{}'.format (linalg.norm (r0_ci_blk.T - ci_blk.T @ ovlp,
            axis=1)))
        if self.ngorb: logger.debug (self, 'Lagrange residual orbital norms '
            'after solving SA-SA part:\n{}'.format (linalg.norm (
                r0_orb)))
        x0 += precond (-r0)
        r1_orb, r1_ci = self.unpack_uniq_var (r0)
        r1_ci_blk = np.asarray ([r1_ci[i].ravel () for i in idx_spin])
        ovlp = ci_blk.conjugate () @ r1_ci_blk.T
        logger.debug (self, 'Lagrange residual SA-SA part after first '
            'precondition:\n{}'.format (ovlp))
        logger.debug (self, 'Lagrange residual SA-CI norms after first '
            'precondition:\n{}'.format (linalg.norm (r1_ci_blk.T - ci_blk.T @ ovlp,
            axis=1)))
        if self.ngorb: logger.debug (self, 'Lagrange residual orbital norms '
            'after first precondition:\n{}'.format (linalg.norm (
                r1_orb)))
        return x0

    def kernel (self, **kwargs):
        '''Cache the effective Hamiltonian terms so you don't have to
        calculate them twice
        '''

        state = kwargs["state"] if "state" in kwargs else self.state
        if state is None:
            raise NotImplementedError("Gradient of PDFT state-average energy")

        self.state = state  # Not the best code hygiene maybe
        mo = kwargs["mo"] if "mo" in kwargs else self.base.mo_coeff
        ci = kwargs["ci"] if "ci" in kwargs else self.base.ci
        if isinstance(ci, np.ndarray):
            ci = [ci]  # hack hack hack...

        kwargs["ci"] = ci
        if ("veff1" not in kwargs) or ("veff2" not in kwargs):
            kwargs["veff1"], kwargs["veff2"] = self.base.get_pdft_veff(
                mo, ci, incl_coul=True, paaa_only=True, state=state, drop_mcwfn=True
            )

        if "mf_grad" not in kwargs:
            kwargs["mf_grad"] = self.base.get_rhf_base().nuc_grad_method()

        return super().kernel (**kwargs)

    def project_Aop (self, Aop, ci, state):
        '''Wrap the Aop function to project out redundant degrees of
        freedom for the CI part.  What's redundant changes between
        SA-CASSCF and MC-PDFT so modify this part in child classes.
        '''
        weights, e_mcscf = np.asarray (self.weights), np.asarray (self.e_mcscf)
        try:
            A_sa = 2 * weights[state] * (e_mcscf - e_mcscf[state])
        except IndexError as e:
            assert (self.nroots == 1), e
            A_sa = 0
        except Exception as e:
            print (self.weights, self.e_mcscf)
            raise (e)
        idx_spin = [i for i in range (self.nroots)
                    if self.spin_states[i]==self.spin_states[state]]
        if self.nroots==1:
            ci_blk = ci.reshape (1, -1)
            ci = [ci]
        else:
            ci_blk = np.asarray ([ci[i].ravel () for i in idx_spin])
            A_sa = A_sa[idx_spin]
        def my_Aop (x):
            Ax = Aop (x)
            x_orb, x_ci = self.unpack_uniq_var (x)
            Ax_orb, Ax_ci = self.unpack_uniq_var (Ax)
            for i, j in product (range (self.nroots), repeat=2):
                if self.spin_states[i] != self.spin_states[j]: continue
                Ax_ci[i] -= np.dot (Ax_ci[i].ravel (), ci[j].ravel ()) * ci[j]
            # Add back in the SA rotation part but from the true energy conds
            x_sa = np.dot (ci_blk.conjugate (), x_ci[state].ravel ())
            Ax_ci[state] += np.dot (x_sa * A_sa, ci_blk).reshape (
                Ax_ci[state].shape)
            Ax = self.pack_uniq_var (Ax_orb, Ax_ci)
            return Ax
        return my_Aop



# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

from pyscf import scf
from pyscf.tdscf import rhf
from pyscf.tdscf import uhf
from pyscf.tdscf import ghf
from pyscf.tdscf import dhf
from pyscf.tdscf.rhf import TDRHF
from pyscf.tdscf.uhf import TDUHF
from pyscf.tdscf.ghf import TDGHF

try:
    from pyscf.dft import KohnShamDFT
    from pyscf.tdscf import rks
    from pyscf.tdscf import uks
    from pyscf.tdscf import gks
    from pyscf.tdscf import dks
    from pyscf.tdscf.rks import TDRKS
    from pyscf.tdscf.uks import TDUKS
    from pyscf.tdscf.gks import TDGKS
except (ImportError, IOError):
    pass


def TDHF(mf, frozen=None):
    if isinstance(mf, scf.hf.KohnShamDFT):
        raise RuntimeError('TDHF does not support DFT object %s' % mf)
    mf = mf.remove_soscf()
    if isinstance(mf, scf.rohf.ROHF):
        # Is it correct to call TDUHF for ROHF?
        mf = mf.to_uhf()
    return mf.TDHF(frozen)

def TDA(mf, frozen=None):
    mf = mf.remove_soscf()
    if isinstance(mf, scf.rohf.ROHF):
        if isinstance(mf, KohnShamDFT):
            mf = mf.to_uks()
        else:
            mf = mf.to_uhf()
    return mf.TDA(frozen)

def TDDFT(mf, frozen=None):
    if isinstance(mf, KohnShamDFT):
        mf = mf.remove_soscf()
        if isinstance(mf, scf.rohf.ROHF):
            mf = mf.to_uks()
        return mf.TDDFT(frozen)
    else:
        return TDHF(mf, frozen)

TD = TDDFT


def RPA(mf, frozen=None):
    return TDDFT(mf, frozen)

def dRPA(mf, frozen=None):
    mf = mf.remove_soscf()
    if isinstance(mf, scf.rohf.ROHF):
        if isinstance(mf, KohnShamDFT):
            mf = mf.to_uks()
        else:
            mf = mf.to_uhf()
    return mf.dRPA(frozen)

def dTDA(mf, frozen=None):
    mf = mf.remove_soscf()
    if isinstance(mf, scf.rohf.ROHF):
        if isinstance(mf, KohnShamDFT):
            mf = mf.to_uks()
        else:
            mf = mf.to_uhf()
    return mf.dTDA(frozen)



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Unrestricted Kohn-Sham for periodic systems with k-point sampling

See Also:
    pyscf.pbc.dft.uks.py : PBC-UKS at a single k-point
'''


import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf.pbc.scf import khf, kuhf
from pyscf.pbc.dft import gen_grid
from pyscf.pbc.dft import rks, krks
from pyscf.pbc.dft.krks import get_rho
from pyscf.pbc.dft import multigrid
from pyscf import __config__


def get_veff(ks, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
             kpts=None, kpts_band=None):
    '''Coulomb + XC functional for UKS.  See pyscf/pbc/dft/uks.py
    :func:`get_veff` fore more details.
    '''
    if cell is None: cell = ks.cell
    if dm is None: dm = ks.make_rdm1()
    if kpts is None: kpts = ks.kpts
    t0 = (logger.process_clock(), logger.perf_counter())

    ni = ks._numint
    if isinstance(ni, multigrid.MultiGridNumInt):
        if ks.do_nlc():
            raise NotImplementedError(f'MultiGrid for NLC functional {ks.xc} + {ks.nlc}')
        j_in_xc = ni.xc_with_j
    else:
        ks.initialize_grids(cell, dm, kpts)
        j_in_xc = False

    max_memory = ks.max_memory - lib.current_memory()[0]
    n, exc, vxc = ni.nr_uks(cell, ks.grids, ks.xc, dm, 0, hermi,
                            kpts, kpts_band, max_memory=max_memory)
    logger.info(ks, 'nelec by numeric integration = %s', n)
    if ks.do_nlc():
        if ni.libxc.is_nlc(ks.xc):
            xc = ks.xc
        else:
            assert ni.libxc.is_nlc(ks.nlc)
            xc = ks.nlc
        n, enlc, vnlc = ni.nr_nlc_vxc(cell, ks.nlcgrids, xc, dm[0]+dm[1],
                                      0, hermi, kpts, max_memory=max_memory)
        exc += enlc
        vxc += vnlc
        logger.info(ks, 'nelec with nlc grids = %s', n)
    t0 = logger.timer(ks, 'vxc', *t0)

    ground_state = kpts_band is None
    nkpts = len(kpts)
    weight = 1. / nkpts
    vj, vk = krks._get_jk(ks, cell, dm, hermi, kpts, kpts_band, with_j=not j_in_xc)
    if j_in_xc:
        ecoul = vxc.ecoul
    else:
        vj = vj[0] + vj[1]
        vxc += vj
        ecoul = None
        if ground_state:
            ecoul = np.einsum('nKij,Kji->', dm, vj).real * .5 * weight
    if ni.libxc.is_hybrid_xc(ks.xc):
        vxc -= vk
        if ground_state:
            exc -= np.einsum('nKij,nKji->', dm, vk).real * .5 * weight
    vxc = lib.tag_array(vxc, ecoul=ecoul, exc=exc, vj=None, vk=None)
    logger.timer(ks, 'veff', *t0)
    return vxc

def energy_elec(mf, dm_kpts=None, h1e_kpts=None, vhf=None):
    if h1e_kpts is None: h1e_kpts = mf.get_hcore(mf.cell, mf.kpts)
    if dm_kpts is None: dm_kpts = mf.make_rdm1()
    if vhf is None or getattr(vhf, 'ecoul', None) is None:
        vhf = mf.get_veff(mf.cell, dm_kpts)

    weight = 1./len(h1e_kpts)
    e1 = weight *(np.einsum('kij,kji', h1e_kpts, dm_kpts[0]) +
                  np.einsum('kij,kji', h1e_kpts, dm_kpts[1]))
    ecoul = vhf.ecoul
    exc = vhf.exc
    tot_e = e1 + ecoul + exc
    mf.scf_summary['e1'] = e1.real
    mf.scf_summary['coul'] = ecoul.real
    mf.scf_summary['exc'] = exc.real
    logger.debug(mf, 'E1 = %s  Ecoul = %s  Exc = %s', e1, ecoul, exc)
    if khf.CHECK_COULOMB_IMAG and abs(ecoul.imag) > mf.cell.precision*10:
        logger.warn(mf, "Coulomb energy has imaginary part %s. "
                    "Coulomb integrals (e-e, e-N) may not converge !",
                    ecoul.imag)
    return tot_e.real, ecoul.real + exc.real

def gen_response(mf, mo_coeff=None, mo_occ=None,
                 with_j=True, hermi=0, max_memory=None, with_nlc=True):
    if mo_coeff is None: mo_coeff = mf.mo_coeff
    if mo_occ is None: mo_occ = mf.mo_occ
    cell = mf.cell
    kpts = mf.kpts
    ni = mf._numint
    hybrid = ni.libxc.is_hybrid_xc(mf.xc)
    j_in_xc = getattr(ni, 'xc_with_j', False)

    if with_nlc and mf.do_nlc():
        raise NotImplementedError

    rho0, vxc, fxc = ni.cache_xc_kernel(cell, mf.grids, mf.xc,
                                        mo_coeff, mo_occ, 1, kpts)
    dm0 = None

    if max_memory is None:
        mem_now = lib.current_memory()[0]
        max_memory = max(2000, mf.max_memory*.8-mem_now)

    def vind(dm1, kshift=0):
        if hermi == 2:
            v1 = np.zeros_like(dm1)
        else:
            assert kshift == 0
            v1 = ni.nr_uks_fxc(cell, mf.grids, mf.xc, dm0, dm1, 0, hermi,
                               rho0, vxc, fxc, kpts, max_memory=max_memory)
        vj, vk = krks._get_jk(mf, cell, dm1, hermi, kpts, with_j=not j_in_xc,
                              kshift=kshift)
        if not j_in_xc:
            v1 += vj[0] + vj[1]
        if hybrid:
            v1 -= vk
        return v1
    return vind

class KUKS(rks.KohnShamDFT, kuhf.KUHF):
    '''UKS class adapted for PBCs with k-point sampling.
    '''

    get_veff = get_veff
    energy_elec = energy_elec
    get_rho = get_rho
    gen_response = gen_response

    def __init__(self, cell, kpts=np.zeros((1,3)), xc='LDA,VWN',
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald')):
        kuhf.KUHF.__init__(self, cell, kpts, exxdiv=exxdiv)
        rks.KohnShamDFT.__init__(self, xc)

    def dump_flags(self, verbose=None):
        kuhf.KUHF.dump_flags(self, verbose)
        rks.KohnShamDFT.dump_flags(self, verbose)
        return self

    def Gradients(self):
        from pyscf.pbc.grad import kuks
        return kuks.Gradients(self)

    def to_hf(self):
        '''Convert to KUHF object.'''
        from pyscf.pbc import scf, df
        out = self._transfer_attrs_(scf.KUHF(self.cell, self.kpts))
        # Pure functionals only construct J-type integrals. Enable all integrals for KHF.
        if (not self._numint.libxc.is_hybrid_xc(self.xc) and
            len(self.kpts) > 1 and getattr(self.with_df, '_j_only', False)):
            out.with_df._j_only = False
            out.with_df.reset()
        return out

    multigrid_numint = krks.KRKS.multigrid_numint

    to_gpu = lib.to_gpu



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Hartree-Fock
============

Simple usage::

    >>> from pyscf import gto, scf
    >>> mol = gto.M(atom='H 0 0 0; H 0 0 1')
    >>> mf = scf.RHF(mol).run()

:func:`scf.RHF` returns an instance of SCF class.  There are some parameters
to control the SCF method.

    verbose : int
        Print level.  Default value equals to :class:`Mole.verbose`
    max_memory : float or int
        Allowed memory in MB.  Default value equals to :class:`Mole.max_memory`
    chkfile : str
        checkpoint file to save MOs, orbital energies etc.
    conv_tol : float
        converge threshold.  Default is 1e-10
    max_cycle : int
        max number of iterations.  Default is 50
    init_guess : str
        initial guess method.  It can be one of 'minao', 'atom', '1e', 'chkfile'.
        Default is 'minao'
    DIIS : class listed in :mod:`scf.diis`
        Default is :class:`diis.SCF_DIIS`. Set it to None/False to turn off DIIS.
    diis : bool
        whether to do DIIS.  Default is True.
    diis_space : int
        DIIS space size.  By default, 8 Fock matrices and errors vector are stored.
    diis_start_cycle : int
        The step to start DIIS.  Default is 0.
    level_shift_factor : float or int
        Level shift (in AU) for virtual space.  Default is 0.
    direct_scf : bool
        Direct SCF is used by default.
    direct_scf_tol : float
        Direct SCF cutoff threshold.  Default is 1e-13.
    callback : function
        callback function takes one dict as the argument which is
        generated by the builtin function :func:`locals`, so that the
        callback function can access all local variables in the current
        environment.
    conv_check : bool
        An extra cycle to check convergence after SCF iterations.

    nelec : (int,int), for UHF/ROHF class
        freeze the number of (alpha,beta) electrons.

    irrep_nelec : dict, for symmetry- RHF/ROHF/UHF class only
        to indicate the number of electrons for each irreps.
        In RHF, give {'ir_name':int, ...} ;
        In ROHF/UHF, give {'ir_name':(int,int), ...} .
        It is effective when :attr:`Mole.symmetry` is set ``True``.

    auxbasis : str, for density fitting SCF only
        Auxiliary basis for density fitting.

        >>> mf = scf.density_fit(scf.UHF(mol))
        >>> mf.scf()

        Density fitting can be applied to all non-relativistic HF class.

    with_ssss : bool, for Dirac-Hartree-Fock only
        If False, ignore small component integrals (SS|SS).  Default is True.
    with_gaunt : bool, for Dirac-Hartree-Fock only
        If False, ignore Gaunt interaction.  Default is False.

Saved results

    converged : bool
        SCF converged or not
    e_tot : float
        Total HF energy (electronic energy plus nuclear repulsion)
    mo_energy :
        Orbital energies
    mo_occ
        Orbital occupancy
    mo_coeff
        Orbital coefficients

'''

from pyscf import gto
from pyscf.scf import hf
rhf = hf
from pyscf.scf import rohf
from pyscf.scf import hf_symm
rhf_symm = hf_symm
from pyscf.scf import uhf
from pyscf.scf import uhf_symm
from pyscf.scf import ghf
from pyscf.scf import ghf_symm
from pyscf.scf import dhf
from pyscf.scf import chkfile
from pyscf.scf import addons
from pyscf.scf import diis
from pyscf.scf import dispersion
from pyscf.scf.diis import DIIS, CDIIS, EDIIS, ADIIS
from pyscf.scf.uhf import spin_square
from pyscf.scf.hf import get_init_guess
from pyscf.scf.addons import *


def HF(mol, *args):
    if mol.nelectron == 1 or mol.spin == 0:
        return RHF(mol, *args)
    else:
        return UHF(mol, *args)
HF.__doc__ = '''
A wrap function to create SCF class (RHF or UHF).\n
''' + hf.SCF.__doc__

def RHF(mol, *args):
    if mol.spin == 0:
        if not mol.symmetry or mol.groupname == 'C1':
            return rhf.RHF(mol, *args)
        else:
            return rhf_symm.RHF(mol, *args)
    else:
        return ROHF(mol, *args)
RHF.__doc__ = hf.RHF.__doc__

def ROHF(mol, *args):
    if mol.nelectron == 1:
        if not mol.symmetry or mol.groupname == 'C1':
            return rohf.HF1e(mol)
        else:
            return hf_symm.HF1e(mol, *args)
    elif not mol.symmetry or mol.groupname == 'C1':
        return rohf.ROHF(mol, *args)
    else:
        return hf_symm.ROHF(mol, *args)
ROHF.__doc__ = rohf.ROHF.__doc__

def UHF(mol, *args):
    if mol.nelectron == 1:
        if not mol.symmetry or mol.groupname == 'C1':
            return uhf.HF1e(mol, *args)
        else:
            return uhf_symm.HF1e(mol, *args)
    elif not mol.symmetry or mol.groupname == 'C1':
        return uhf.UHF(mol, *args)
    else:
        return uhf_symm.UHF(mol, *args)
UHF.__doc__ = uhf.UHF.__doc__

def GHF(mol, *args):
    if mol.nelectron == 1:
        if not mol.symmetry or mol.groupname == 'C1':
            return ghf.HF1e(mol)
        else:
            return ghf_symm.HF1e(mol, *args)
    elif not mol.symmetry or mol.groupname == 'C1':
        return ghf.GHF(mol, *args)
    else:
        return ghf_symm.GHF(mol, *args)
GHF.__doc__ = ghf.GHF.__doc__

def DHF(mol, *args):
    if mol.nelectron == 1:
        return dhf.HF1e(mol)
    elif dhf.zquatev and mol.spin == 0:
        return dhf.RDHF(mol, *args)
    else:
        return dhf.DHF(mol, *args)
DHF.__doc__ = dhf.DHF.__doc__


def X2C(mol, *args):
    '''X2C Hartree-Fock'''
    from pyscf.x2c import x2c
    if dhf.zquatev and mol.spin == 0:
        return x2c.RHF(mol, *args)
    else:
        return x2c.UHF(mol, *args)
X2C_HF = X2C

def sfx2c1e(mf):
    '''spin-free (the scalar part) X2C with 1-electron X-matrix'''
    return mf.sfx2c1e()
sfx2c = sfx2c1e

def density_fit(mf, auxbasis=None, with_df=None, only_dfj=False):
    return mf.density_fit(auxbasis, with_df, only_dfj)

def newton(mf):
    from pyscf.soscf import newton_ah
    return newton_ah.newton(mf)

fast_newton = addons.fast_newton

def KS(mol, *args):
    from pyscf import dft
    return dft.KS(mol, *args)

def RKS(mol, *args):
    from pyscf import dft
    return dft.RKS(mol, *args)

def ROKS(mol, *args):
    from pyscf import dft
    return dft.ROKS(mol, *args)

def UKS(mol, *args):
    from pyscf import dft
    return dft.UKS(mol, *args)

def GKS(mol, *args):
    from pyscf import dft
    return dft.GKS(mol, *args)

def DKS(mol, *args):
    from pyscf import dft
    return dft.DKS(mol, *args)
